/* $Id: gps.c,v 1.17 2012-03-11 17:25:24 simimeie Exp $
 * Functions for handling the GPS receiver 
 * (and the time reception done through it)
 */

#if ((defined(GPS)) && (defined(SERIALCONSOLE)))
#error "GPS and Serial Console are incompatible as they use the same pins."
#endif

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <util/delay.h>
#include <ctype.h>
#include "gps.h"
#include "console.h"
#include "timeconvert.h"

#if (defined(GPS))

/* The GPS receiver defaults to 9600 baud */
#define UBRRCALC ((CPUFREQ / (16 * 9600UL)) - 1)

#define GPSBUFSIZE 200
static uint8_t gpsbuf[GPSBUFSIZE+1];
static uint16_t gpsbpos = 0;
static volatile uint8_t lastskytraqmsgack = 0;
static uint8_t lastskytraqmsgnum = 0;
static volatile struct timestamp tsoflastpps;
static uint8_t msgssincelastgsv = 0; /* GPS messages since the last GPGSV message */

/* these are globally exported */
struct gpstime gpstime[2];
volatile uint8_t gpsvalid;
uint8_t gpssatsinview = 0;

/* Returns: 1 if checksum is correct, 0 otherwise */
static uint8_t gps_checknmeachecksum(void)
{
  uint16_t i = 1;
  uint8_t csc = 0; /* calculated checksum */
  uint8_t csr; /* right checksum */
  if (gpsbuf[0] != '$') {
    /* Not an NMEA message */
    return 0;
  }
  while ((gpsbuf[i] != 0) && (gpsbuf[i] != '*')) {
    csc ^= gpsbuf[i];
    i++;
  }
  if (gpsbuf[i] != '*') { /* No end marker -> invalid */
    return 0;
  }
  i++;
  switch (gpsbuf[i]) {
  case '0' ... '9':  csr = ((gpsbuf[i] - '0') +  0) << 4; break;
  case 'A' ... 'F':  csr = ((gpsbuf[i] - 'A') + 10) << 4; break;
  case 'a' ... 'f':  csr = ((gpsbuf[i] - 'a') + 10)<< 4; break;
  default: return 0;
  };
  i++;
  switch (gpsbuf[i]) {
  case '0' ... '9':  csr |= ((gpsbuf[i] - '0') +  0) << 0; break;
  case 'A' ... 'F':  csr |= ((gpsbuf[i] - 'A') + 10) << 0; break;
  case 'a' ... 'f':  csr |= ((gpsbuf[i] - 'a') + 10) << 0; break;
  default: return 0;
  };
  i++;
  if (gpsbuf[i] != '\r') { /* Not the end of the string? */
    return 0;
  }
  if (csc != csr) { /* Checksum does not match */
#if 0
    console_printpgm_P(PSTR("\r\n"));
    console_printhex8(csc);
    console_printpgm_P(PSTR(" <-> "));
    console_printhex8(csr);
    console_printpgm_P(PSTR("\r\n"));
#endif
    return 0;
  }
  return 1;
}

static void gps_parsenmeamessage(void)
{
  if (gpsbpos < 22) {
    return; /* too short */
  }
  if (gps_checknmeachecksum() != 1) { /* CRC failure */
    /* Ignore the message */
    return;
  }
#ifdef NOGPSZDA
  /* As it turns out, GPRMC is not the best for the job. That is because
   * the message is usually not sent synchronously, i.e. it is sent at a
   * random point during the second, and receivers tend to deliver them
   * with random delays also, making a reliable mapping of PPS pulse to
   * corresponding RMC message impossible. If your receiver can output
   * GPZDA (it's a feature that has to be enabled by the GPS receiver
   * manufacturer when compiling the firmware), this is definitely what
   * you want to use instead.
   * So here's the fallback code for GPRMC. */
  if ((gpsbuf[1] == 'G') && (gpsbuf[2] == 'P') && (gpsbuf[3] == 'R')
   && (gpsbuf[4] == 'M') && (gpsbuf[5] == 'C') && (gpsbuf[6] == ',')) {
      uint8_t ppos = 19;
      uint8_t numcommas = 0;
      struct brokentime brt;
      struct timestamp curts;
      uint32_t tsdif;
      uint8_t nxtact = (gpsvalid ? 0 : 1);
      /* Example message:
       *        v Time (HHMMSS.SSS)                               v Date (DDMMYY)
       * $GPRMC,124545.000,V,0000.0000,N,00000.0000,E,000.0,000.0,280606,,,N*7F */
      brt.hour = (gpsbuf[7] - '0') * 10 + (gpsbuf[8] - '0');
      brt.min = (gpsbuf[9] - '0') * 10 + (gpsbuf[10] - '0');
      brt.sec = (gpsbuf[11] - '0') * 10 + (gpsbuf[12] - '0');
      if ((gpsbuf[13] != '.') || (gpsbuf[17] != ',') || (gpsbuf[18] != 'A')) {
        /* Data not formatted correctly or not marked as valid
         * (at Pos. 18: A = valid, V = invalid) */
        return;
      }
      while ((gpsbuf[ppos] != '*') && (gpsbuf[ppos] != 0)) {
        if (gpsbuf[ppos] == ',') {
          numcommas++;
        }
        if (numcommas == 7) { break; }
        ppos++;
      }
      /* We're now at the end of the string or the date field */
      if (gpsbuf[ppos] != ',') {
        /* End of string -> parsing failed -> abort */
        return;
      }
      ppos++;
      brt.day = (gpsbuf[ppos] - '0') * 10 + (gpsbuf[ppos + 1] - '0');
      ppos += 2;
      brt.month = (gpsbuf[ppos] - '0') * 10 + (gpsbuf[ppos + 1] - '0');
      ppos += 2;
      brt.year = (gpsbuf[ppos] - '0') * 10 + (gpsbuf[ppos + 1] - '0');
      /* check sanity (less than 1 second since tsoflastpps) */
      gettimestamp_noirq(&curts);
      tsdif = (((uint32_t)curts.tovfls << 16) | curts.ticks)
            - (((uint32_t)tsoflastpps.tovfls << 16) | tsoflastpps.ticks);
      if (tsdif < TICKSPERSECOND) { /* valid! */
        gpstime[nxtact].unixts = bttots(&brt);
        gpstime[nxtact].devicetime = tsoflastpps;
        gpstime[nxtact].issummertime = isdst(gpstime[nxtact].unixts, brt.year);
        gpsvalid = nxtact;
      }
      if (0) {
        uint8_t spfbuf[30];
        console_printpgm_noirq_P(PSTR(" PD:"));
        sprintf(spfbuf, "%02d.%02d.20%02d %02d:%02d:%02d ",
                        brt.day, brt.month, brt.year,
                        brt.hour, brt.min, brt.sec);
        console_printtext_noirq(spfbuf);
        console_printpgm_noirq_P(PSTR("\r\n"));
      }
      return;
  }
#endif /* NOGPSZDA */
#ifndef NOGPSZDA
  /* Yay, we have GPZDA available. As this message is actually made for
   * timekeeping purposes, it is of course perfect for us.
   * Unfortunately, it seems that it's not sent synchronously on our
   * venus638, so we need a workaround. */
  if ((gpsbuf[1] == 'G') && (gpsbuf[2] == 'P') && (gpsbuf[3] == 'Z')
   && (gpsbuf[4] == 'D') && (gpsbuf[5] == 'A') && (gpsbuf[6] == ',')) {
      uint8_t ppos = 13;
      uint8_t numcommas = 0;
      struct brokentime brt;
      struct timestamp curts;
      uint32_t tsdif;
      uint8_t nxtact = (gpsvalid ? 0 : 1);
      /* Example message:
       *  $GPZDA,192346.473,09,03,2012,00,00*56
       *  $GPZDA,Time (HHMMSS.SSS),Day,Month,Year,
       */
      brt.hour = (gpsbuf[7] - '0') * 10 + (gpsbuf[8] - '0');
      brt.min = (gpsbuf[9] - '0') * 10 + (gpsbuf[10] - '0');
      brt.sec = (gpsbuf[11] - '0') * 10 + (gpsbuf[12] - '0');
      /* Sometimes there is no dot at all, sometimes there are 2 digits,
       * sometimes 3. We'll try to be tolerant here. */
      while ((gpsbuf[ppos] != '*') && (gpsbuf[ppos] != 0)) {
        if (gpsbuf[ppos] == ',') {
          numcommas++;
          if (numcommas == 2) { /* The previous field was the day */
            brt.day = (gpsbuf[ppos - 2] - '0') * 10 + (gpsbuf[ppos - 1] - '0');
          }
          if (numcommas == 3) { /* The previous field was the month */
            brt.month = (gpsbuf[ppos - 2] - '0') * 10 + (gpsbuf[ppos - 1] - '0');
          }
          if (numcommas == 4) { /* The previous field was the year */
            brt.year = (gpsbuf[ppos - 2] - '0') * 10 + (gpsbuf[ppos - 1] - '0');
            if (brt.year < 10) { /* Before 2010? Nonsense. It's 2012 as I'm writing this. */
              return;
            }
          }
        }
        ppos++;
      }
      /* check sanity (less than 1 second since tsoflastpps) */
      gettimestamp_noirq(&curts);
      tsdif = (((uint32_t)curts.tovfls << 16) | curts.ticks)
            - (((uint32_t)tsoflastpps.tovfls << 16) | tsoflastpps.ticks);
      if (tsdif < TICKSPERSECOND) { /* valid! */
        gpstime[nxtact].unixts = bttots(&brt);
        if (gpsbuf[13] == '.') { /* the time was actually given with high accuracy (HHMMSS.XX) */
          if (gpsbuf[14] >= '7') {
            /* If there are high values for the fraction, we're at risk of
             * mistakenly matching the PPS pulse to the wrong second. */
            if (tsdif < (TICKSPERSECOND / 3)) {
              /* So the pulse was less than .3 seconds ago, but the timestamp
               * in the message says .[789]. That means the pulse was actually
               * for the second AFTER the message! */
              gpstime[nxtact].unixts++;
              /* console_printpgm_noirq_P(PSTR(" PCOR ")); */
            }
          }
        }
        gpstime[nxtact].devicetime = tsoflastpps;
        gpstime[nxtact].issummertime = isdst(gpstime[nxtact].unixts, brt.year);
        gpsvalid = nxtact;
      }
      if (0) {
        uint8_t spfbuf[30];
        console_printpgm_noirq_P(PSTR(" PD:"));
        sprintf(spfbuf, "%02d.%02d.20%02d %02d:%02d:%02d ",
                        brt.day, brt.month, brt.year,
                        brt.hour, brt.min, brt.sec);
        console_printtext_noirq(spfbuf);
        console_printpgm_noirq_P(PSTR("\r\n"));
      }
      return;
  }
#endif /* not NOGPSZDA */
  /* If no satellites are in view at all, we do not get any GPGSV messages.
   * We thus need some sort of timeout mechanism to reset gpssatsinview to
   * zero when that happens. */
  msgssincelastgsv++;
  if (msgssincelastgsv > 10) {
    gpssatsinview = 0;
  }
  if ((gpsbuf[1] == 'G') && (gpsbuf[2] == 'P') && (gpsbuf[3] == 'G')
   && (gpsbuf[4] == 'S') && (gpsbuf[5] == 'V') && (gpsbuf[6] == ',')) {
      /* Satellites in view
       * Example message:
       * $GPGSV,1,1,01,13,00,000,27*4F
       *    */
      uint8_t ppos = 5;
      uint8_t numcommas = 0;
      while ((gpsbuf[ppos] != '*') && (gpsbuf[ppos] != 0)) {
        if (gpsbuf[ppos] == ',') {
          numcommas++;
        }
        if (numcommas == 3) { break; }
        ppos++;
      }
      /* We're now at the end of the string or in the right field */
      if (gpsbuf[ppos] != ',') {
        /* End of string -> parsing failed -> abort */
        return;
      }
      gpssatsinview = (gpsbuf[ppos + 1] - '0') * 10 + (gpsbuf[ppos + 2] - '0');
      msgssincelastgsv = 0;
      return;
  }
#if 0 /* this is useless info - use GPGSV instead. */
  if ((gpsbuf[1] == 'G') && (gpsbuf[2] == 'P') && (gpsbuf[3] == 'G')
   && (gpsbuf[4] == 'S') && (gpsbuf[5] == 'A') && (gpsbuf[6] == ',')) {
      /* Some general information
       * Example message:
       * $GPGSA,A,1,,,,,,,,,,,,,6.0,5.3,2.8*3A
       *    */
      uint8_t ppos = 11;
      uint8_t numcommas = 0;
      gpssatsused = 0;
      if ((gpsbuf[8] != ',') || (gpsbuf[9] == '1') || (gpsbuf[10] != ',')) {
        /* Format error, or mode2 is '1' - which means 'no fix available' */
        return;
      }
      while ((gpsbuf[ppos] != '*') && (gpsbuf[ppos] != 0)) {
        if (gpsbuf[ppos] == ',') {
          numcommas++;
        }
        if (numcommas >= 12) { break; /* only the first 12 are sats */ }
        if (isdigit(gpsbuf[ppos])) {
          gpssatsused++;
          while (isdigit(gpsbuf[ppos])) {
            ppos++;
          }
        } else { /* anything that is not a number - jump to next */
          ppos++;
        }
      }
  }
#endif
}

static void gps_parsebinarymessage(void)
{
  if (gpsbuf[4] == 0x83) {
    /* console_printpgm_P(PSTR(" ACKsc"));
    console_printhex8(gpsbuf[5]); */
    if (gpsbuf[5] == lastskytraqmsgnum) {
      lastskytraqmsgack = 1;
    }
/*  } else if (gpsbuf[4] == 0x84) {
    console_printpgm_P(PSTR(" NAKsc"));
    console_printhex8(gpsbuf[5]);
  } else {
    console_printpgm_P(PSTR(" UnknownSkytraqBin"));
    console_printhex8(gpsbuf[4]); */
  }
  /* console_printpgm_P(PSTR(" ")); */
}

/* Handler for RXC (RX Complete) IRQ. */
ISR(USART_RX_vect)
{
  uint8_t inpb;
  
  inpb = UDR0;
#if 0
  if ((inpb < '\n') || (inpb >= 0x80)) {
    console_printchar_noirq('x');
    console_printhex8_noirq(inpb);
  } else {
    console_printchar_noirq(inpb);
  }
#endif
  if (inpb == '\n') { /* End of line */
    /* xA0xA1x00x02x83x00x83
       xA0xA1x00x02x83x08x8B */
    if ((gpsbpos >= 2) && (gpsbuf[0] == 0xA0) && (gpsbuf[1] == 0xA1)) {
      /* Seems to be an skytraq binary message */
      uint16_t l = 999; /* Default to a large value as long as we don't know */
      if (gpsbpos > 4) { /* Now that we have length information, use it */
        l = ((uint16_t)gpsbuf[2] << 8) + gpsbuf[3] + 4 + 2;
      }
      if (l >= GPSBUFSIZE) {
        l = GPSBUFSIZE - 1;
      }
      if (gpsbpos < l) { /* We're not at the end of the binary message yet, */
                         /* it's just a coincidence it contained the \n */
        gpsbuf[gpsbpos] = inpb;
        gpsbpos++;
      } else {
        gps_parsebinarymessage();
        gpsbpos = 0;
      }
    } else {
      /* probably an NMEA message */
      gpsbuf[gpsbpos] = 0;
      gps_parsenmeamessage();
      gpsbpos = 0;
    }
  } else {
    if (gpsbpos < GPSBUFSIZE) {
      gpsbuf[gpsbpos] = inpb;
      gpsbpos++;
    }
  }
}

/* Handler for PPS (Pulse Per Second) IRQ */
ISR(INT0_vect)
{
  gettimestamp_noirq((struct timestamp *)&tsoflastpps);
#if 0
  console_printpgm_noirq_P(PSTR(" PPS"));
  if ((PIND & _BV(2))) {
    console_printpgm_noirq_P(PSTR("1 "));
  } else {
    console_printpgm_noirq_P(PSTR("0 "));
  }
#endif
}

#define gps_sendonebyte(x) \
  while ((UCSR0A & _BV(UDRE0)) == 0) { /* Not ready yet for next byte, wait */ } \
  UDR0 = (x);

static void gps_sendtogpsskytraqbin(uint16_t l, PGM_P what)
{
  uint8_t t;
  uint8_t cs = 0;
  gps_sendonebyte(0xA0);
  gps_sendonebyte(0xA1);
  gps_sendonebyte(l >> 8);
  gps_sendonebyte(l & 0xff);
  /* Keep track of this message so we can see if it gets acked. */
  t = pgm_read_byte(what);
  lastskytraqmsgack = 0; lastskytraqmsgnum = t;
  while (l > 0) {
    t = pgm_read_byte(what++);
    gps_sendonebyte(t);
    cs ^= t;
    l--;
  }
  gps_sendonebyte(cs);
  gps_sendonebyte(0x0d);
  gps_sendonebyte(0x0a);
}

void gps_init(void)
{
  /* Set Baud Rate */
  UBRR0H = (uint8_t)(UBRRCALC >> 8);
  UBRR0L = (uint8_t)(UBRRCALC);
  /* Set 8 Bit mode - how braindead is it to default to 5 bit?! */
  UCSR0C = /* _BV(URSEL) | */ _BV(UCSZ00) | _BV(UCSZ01);
  /* Enable Reception and Receive IRQ */
  UCSR0B = _BV(RXEN0) | _BV(RXCIE0) | _BV(TXEN0);
  /* Enable the PPS IRQ (IRQ0) and let it trigger on the rising edge. */
  EICRA |= _BV(ISC00) | _BV(ISC01);
  EIMSK |= _BV(INT0);
}

uint8_t gps_configurereceiver(void)
{
  uint8_t wcnt;
  /* skytraq binary commands */
  /* Set NMEA message output.
   * Each parameter is for one NMEA message and sets every how many seconds
   * that message is repeated. 0 disables the message. 
   * (0) \x08 is the command, what follows are the values for:
   * (1) GGA, (2) GSA, (3) GSV, (4) GLL, (5) RMC, (6) VTG, (7) ZDA */
  /* If your GPS receiver does not support ZDA, then RMC can be used as a
   * (rather crappy) replacement. */
#ifdef NOGPSZDA
  /*                                    GGA GSA GSV GLL RMC VTG ZDA */
  gps_sendtogpsskytraqbin(9, PSTR("\x08\x00\x00\x05\x00\x01\x00\x00\x00"));
#else /* we do have the GPZDA message from the GPS */
  /*                                    GGA GSA GSV GLL RMC VTG ZDA */
  gps_sendtogpsskytraqbin(9, PSTR("\x08\x00\x00\x05\x00\x00\x00\x01\x00"));
#endif /* NOGPSZDA */
  wcnt = 0;
  while ((lastskytraqmsgack != 1) && (wcnt < 0xff)) {
    _delay_ms(1.5);
    wcnt++;
  }
  if (wcnt == 0xff) { /* Failed due to timeout */
    return 0;
  }
  return 1;
}

#else /* !GPS */

/* define empty dummy functions */

void gps_init(void) { }

uint8_t gps_configurereceiver(void) {
  return 0;
}

#endif /* GPS */
