/* $Id: network.c,v 1.16 2012-03-10 13:08:12 simimeie Exp $
 * Network functions.
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/version.h>
#include <util/delay.h>
#include <stdio.h>
#include <string.h>
#include "network.h"
#include "console.h"
#include "timers.h"
#include "gps.h"
#include "ledmodule.h"
#include "rcondefs.h"

/* Some Settings */

/* Number of TX buffers to allocate in the memory of the ENC28J60.
 * Note that the higher you set this, the more packets you can have
 * pending to send, but the less buffer memory is available for incoming
 * packets. Each TX buffer eats away 1526 bytes of the ENC28J60s 8192. */
#define NUMTXBUFS 3

/* Maximum Number of open TCP connections. Note that each one needs
 * about 40 Bytes of memory. */
#define NUMTCPCONNS 4

/* On which port do we accept telnet connections?
 * (if NETCONSOLE is defined) */
#define TELNETPORT 23

/* What TTL to set on outgoing packets? */
#define NET_TTL 0x40

/* How many seconds of inactivity before we time out a TCP connection?
 * Note: This is by design limited to no more than 300 seconds (or even
 * less if we're clocked with more than 12 MHz) */
#define TCP_TIMEOUT 60

/*************************************************************************
 * Everything below this line should only be touched by people that know *
 * what they're doing - and me.                                          *
 *********************************************************************** */
/* ----------------------------- this line ----------------------------- */

/* MAC address */
uint8_t net_mac[6];

/* Full Duplex enabled */
uint8_t net_fdx;

/* IPv4 address */
uint8_t net_ip4[4];

/* Global IPv6 address */
uint8_t net_ip6[16];

/* Link-Local V6 address (always generated automatically) */
static uint8_t net_llip6[16];

/* IPv6 addressing mode. See eepromdata.h for the meaning of the values. */
uint8_t net_ipv6addrmode;

/* Remote CONtrol password */
uint8_t net_rconpass[16];

static uint16_t currxad;
static uint16_t curtxad;

static uint8_t txbufused[NUMTXBUFS];
static uint16_t txbufstart[NUMTXBUFS];
static uint16_t txbufpsize[NUMTXBUFS];
static uint8_t txsending;

#define STATE_SYNACKSENT 1
#define STATE_ESTABLISHED 2
#define STATE_LASTACK 3
struct tcpstate {  /* This state can be used for both IPv4 and IPv6 connections */
  union {
    uint8_t partnerip4[4];
    uint8_t partnerip6[16];
  };
  uint16_t lastactive; /* Timestamp (in ticks) when this connection was
                        * last active (=receiving an ACK) */
  uint32_t partnersnr;
  uint32_t mysnr;
  uint16_t partnerport;
  uint16_t myport;
  uint8_t partnermac[6];
  uint8_t iptype:1,  /* 0: IPv4, 1: IPv6 */
          state:4;   /* See STATE definitions above */
};

struct tcpstate tcpconns[NUMTCPCONNS];

/* (Reused) buffer for headers of incoming packet. The first n Bytes
 * of every incoming packet are copied to the header buffer for easier
 * parsing. You should usually NOT touch this except for debugging. */
#define HEADERBUFSIZE 84
static uint8_t hdrbuf[HEADERBUFSIZE];
/* start of the current packet in the enc28j60 memory */
static uint16_t thispacketstart;
/* length of the current packet */
static uint16_t plen;

/* Bit set for multicast packets in the first byte of the MAC address */
#define NET_MCASTBIT       0x01

/* Some special IPv6 addresses */
static PROGMEM uint8_t ipv6_solicitednode[13] = { 0xFF, 0x02, 0x00, 0x00,
                                                  0x00, 0x00, 0x00, 0x00,
                                                  0x00, 0x00, 0x00, 0x01,
                                                  0xFF /* last 3 bytes ignored */ };

static PROGMEM uint8_t ipv6_allnodes[16]      = { 0xFF, 0x02, 0x00, 0x00,
                                                  0x00, 0x00, 0x00, 0x00,
                                                  0x00, 0x00, 0x00, 0x00,
                                                  0x00, 0x00, 0x00, 0x01 };

#ifndef MINIMUM
#define MINIMUM(x, y) (((x) < (y)) ? (x) : (y))
#endif /* MINIMUM */

#ifndef SHOWNETERRORS
#define SHOWNETERRORS 0
#endif

/* This is intended to be used with some more important messages that
 * might be interesting even after debugging */
#define ERRORconsole_pgm_P(x) \
        if (SHOWNETERRORS) { \
          console_printpgm_P(x); \
        }

/* ENC28J60 chip defines */
/* Adresses mapped out */
#define ADDR_MASK 0x1F
/* Is this register a MAC or MII register? This is relevant, because then
 * we get an additional dummy byte on a read. */
#define ISMACMII_MASK 0x80
/* Which bank is the register in? */
#define BANKOF(x) ((x & 0x60) >> 5)

/* Our receive buffer goes from 0 up to and including this address */
#define RXBUFEND (0x1fff - (NUMTXBUFS * 1526)) /* 1518 ethernet frame + 7 for internal chip use */

/* *********************************************************************
 * Register definitions. don't you just love them.
 ********************************************************************* */
/* Memory read pointer */
#define E28J60_ERDPTL   0x00
#define E28J60_ERDPTH   0x01
/* Memory write pointer */
#define E28J60_EWRPTL   0x02
#define E28J60_EWRPTH   0x03
/* TX Start */
#define E28J60_ETXSTL   0x04
#define E28J60_ETXSTH   0x05
/* TX End */
#define E28J60_ETXNDL   0x06
#define E28J60_ETXNDH   0x07
/* RX Start */
#define E28J60_ERXSTL   0x08
#define E28J60_ERXSTH   0x09
/* RX End */
#define E28J60_ERXNDL   0x0a
#define E28J60_ERXNDH   0x0b
/* RX Read Pointer */
#define E28J60_ERXRDPTL 0x0c
#define E28J60_ERXRDPTH 0x0d
/* RX Write Pointer */
#define E28J60_ERXWRPTL 0x0e
#define E28J60_ERXWRPTH 0x0f
/* EIE: Interrupt control */
#define E28J60_EIE      0x1b
#define E28J60_EIE_INTIE  0x80 /* global interrupt enable */
#define E28J60_EIE_PKTIE  0x40 /* receive pending packet int enable */
#define E28J60_EIE_DMAIE  0x20 /* DMA int enable */
#define E28J60_EIE_LINKIE 0x10 /* link change int enable */
#define E28J60_EIE_TXIE   0x08 /* TX int enable */
#define E28J60_EIE_TXERIE 0x02 /* TX error int enable */
#define E28J60_EIE_RXERIE 0x01 /* RX error int enable */
/* ESTAT */
#define E28J60_ESTAT    0x1d
/* ECON control registers */
#define E28J60_ECON2    0x1e
#define E28J60_ECON2_AUTOINC  0x80 /* Automatic Buffer Pointer Increment */
#define E28J60_ECON2_PKTDEC   0x40 /* decrement packet counter */
#define E28J60_ECON1    0x1f
#define E28J60_ECON1_TXRTS 0x08  /* Transmit Request to Send */
#define E28J60_ECON1_RXEN  0x04  /* RX enable */

/* Receive filter register */
#define E28J60_ERXFCON  0x38
#define E28J60_ERXFCON_UCEN   0x80   /* Unicast */
#define E28J60_ERXFCON_ANDOR  0x40   /* AND/OR select (1 = AND, 0 = OR) */
#define E28J60_ERXFCON_CRCEN  0x20   /* CRC checks */
#define E28J60_ERXFCON_PMEN   0x10   /* pattern Match filter */
#define E28J60_ERXFCON_MPEN   0x08   /* magic packet filter */
#define E28J60_ERXFCON_HTEN   0x04   /* hash table filter */
#define E28J60_ERXFCON_MCEN   0x02   /* multicast filter */
#define E28J60_ERXFCON_BCEN   0x01   /* broadcast filter */
/* Number of received packets still unhandled */
#define E28J60_EPKTCNT  0x39

/* MAC control */
#define E28J60_MACON1   0xc0
#define E28J60_MACON1_MARXEN   0x01   /* RX Enable */
#define E28J60_MACON1_PASSALL  0x02   /* Pass All Received Frames (incl. control frames) */
#define E28J60_MACON1_RXPAUS   0x04   /* Allow the MAC to react to pause control frames */
#define E28J60_MACON1_TXPAUS   0x08   /* Allow the MAC to transmit pause control frames (needed for FDX) */
/* This register is NOT documented at all in the current version of the
 * documentation? In fact, it's listed as reserved? */
#define E28J60_MACON2   0xc1
#define E28J60_MACON2_MARST     0x80
#define E28J60_MACON2_RNDRST    0x40
#define E28J60_MACON2_MARXRST   0x08
#define E28J60_MACON2_RFUNRST   0x04
#define E28J60_MACON2_MATXRST   0x02
#define E28J60_MACON2_TFUNRST   0x01
#define E28J60_MACON3   0xc2
#define E28J60_MACON3_PADCFG2  0x80   /* Automatic Pad and CRC Configuration (2) */
#define E28J60_MACON3_PADCFG1  0x40   /* Automatic Pad and CRC Configuration (2) */
#define E28J60_MACON3_PADCFG0  0x20   /* Automatic Pad and CRC Configuration (2) */
#define E28J60_MACON3_TXCRCEN  0x10   /* transmit CRC enable */
#define E28J60_MACON3_FULDPX   0x01   /* Full duplex enable */
#define E28J60_MACON4   0xc3
#define E28J60_MACON4_DEFER   0x40    /* Defer Transmission Enable */
#define E28J60_MACON4_BPEN    0x20    /* No Backoff during Backpressure Enable */
#define E28J60_MACON4_NOBKOFF 0x10    /* No Backoff Enable */
/* Back-to-back inter-packet gap */
#define E28J60_MABBIPG  0xc4
/* Non-Back-to-back inter-packet gap */
#define E28J60_MAIPGL   0xc6
#define E28J60_MAIPGH   0xc7
/* Maximum frame length */
#define E28J60_MAMXFLL  0xca
#define E28J60_MAMXFLH  0xcb
/* MII register address */
#define E28J60_MIREGADR 0xd4
/* MII write */
#define E28J60_MIWRL    0xd6
#define E28J60_MIWRH    0xd7

/* MAC address */
#define E28J60_MAADR1   0xe4
#define E28J60_MAADR2   0xe5
#define E28J60_MAADR3   0xe2
#define E28J60_MAADR4   0xe3
#define E28J60_MAADR5   0xe0
#define E28J60_MAADR6   0xe1
/* MII status */
#define E28J60_MISTAT   0xea
#define E28J60_MISTAT_NVALID   0x04  /* contents of MIRD not valid yet */
#define E28J60_MISTAT_SCAN     0x02  /* MII management scan operation */
#define E28J60_MISTAT_BUSY     0x01  /* MII management busy (r or w in progress) */
/* chip Revision info */
#define E28J60_EREVID   0x72
/* clock output control */
#define E28J60_ECOCON   0x75

/* PHY defines. These adresses have nothing to do with the ones
 * above! */
#define E28J60_PHCON1   0x00 /* PHY control 1 */
#define E28J60_PHCON1_PDPXMD     0x0100  /* Full Duplex */
#define E28J60_PHSTAT1  0x01 /* PHY stat 1 */
#define E28J60_PHCON2   0x10 /* PHY control 2 */
#define E28J60_PHCON2_FRCLNK     0x4000  /* Force Link Up */
#define E28J60_PHCON2_TXDIS      0x2000  /* twisted pair transmitter disable */
#define E28J60_PHCON2_HDLDIS     0x0100  /* half duplex loopback disable */
#define E28J60_PHSTAT2  0x11 /* PHY stat 2 */
#define E28J60_PHIE     0x12 /* PHY ? */
#define E28J60_PHIR     0x13 /* PHY ? */
#define E28J60_PHLCON   0x14 /* PHY LED control */

/* And a few TCP defines */
#define TCP_RELEVANTFLAGMASK 0x17
#define TCP_ACK 0x10
#define TCP_PSH 0x08
#define TCP_RST 0x04
#define TCP_SYN 0x02
#define TCP_FIN 0x01

/* Currently selected enc28j60 page */
static uint8_t curbank = 0;

/* Note: the signal is inverted, thus pulling this high DEselects the chip */
#define spi_ss_hi() { PORTB |= _BV(PB2); }
/* Consequently, this SELECTS the chip */
#define spi_ss_lo() { PORTB &= (uint8_t)~_BV(PB2); }

static void spi_sendbyte(uint8_t v) {
  /* Start transmission */
  SPDR = v;
  /* Wait for transmission complete */
  while ((SPSR & _BV(SPIF)) == 0) {
  }
}

static uint8_t spi_readbyte() {
  SPDR = 0x00; /* Dummy write */
  while ((SPSR & _BV(SPIF)) == 0) {
  }
  /* Return Data Register */
  return SPDR;
}

/* This gets called recursively. There is no endless loop, because the
 * ad >= 0x1a aborting condition in setbank is always true if called
 * from there */
static void enc28j60_setbank(uint8_t ad);

static void enc28j60_bitfield_set(uint8_t ad, uint8_t val) {
  enc28j60_setbank(ad);
  spi_ss_lo();
  spi_sendbyte(0x80 | (ad & ADDR_MASK));
  spi_sendbyte(val);
  spi_ss_hi();
}

static void enc28j60_bitfield_clear(uint8_t ad, uint8_t val) {
  enc28j60_setbank(ad);
  spi_ss_lo();
  spi_sendbyte(0xa0 | (ad & ADDR_MASK));
  spi_sendbyte(val);
  spi_ss_hi();
}

static void enc28j60_setbank(uint8_t ad) {
  uint8_t bank = BANKOF(ad);
  
  if (bank == curbank) { /* already the correct bank active */
    return;
  }
  if ((ad & ADDR_MASK) >= 0x1a) { /* No need to switch banks for the common registers */
    return;
  }
  /*console_printpgm_P(PSTR(" BSW "));
  console_printhex8(bank);
  console_printpgm_P(PSTR(" ")); */
  if (bank & (1 << 0)) {
    enc28j60_bitfield_set(E28J60_ECON1, 1 << 0);
  } else {
    enc28j60_bitfield_clear(E28J60_ECON1, 1 << 0);
  }
  if (bank & (1 << 1)) {
    enc28j60_bitfield_set(E28J60_ECON1, 1 << 1);
  } else {
    enc28j60_bitfield_clear(E28J60_ECON1, 1 << 1);
  }
  curbank = bank;
}

static uint8_t enc28j60_read_cont_reg(uint8_t ad) {
  uint8_t res;
  enc28j60_setbank(ad);
  spi_ss_lo();
  spi_sendbyte(0x00 | (ad & ADDR_MASK));
  if (ad & 0x80) {
    /* We get a dummy byte first that we need to ignore */
    spi_readbyte();
  }
  res = spi_readbyte();
  spi_ss_hi();
  return res;
}

/* Just for debugging: Export our internal function */
uint8_t enc28j60_readregister(uint8_t ad) {
  return enc28j60_read_cont_reg(ad);
}

static void enc28j60_write_cont_reg(uint8_t ad, uint8_t v) {
  enc28j60_setbank(ad);
  spi_ss_lo();
  spi_sendbyte(0x40 | (ad & ADDR_MASK));
  spi_sendbyte(v);
  spi_ss_hi();
}

static uint8_t enc28j60_read_memory(uint16_t ad) {
  uint8_t res;
  
  ad &= 0x1fff;
  if (ad != currxad) { /* move pointer */
    enc28j60_write_cont_reg(E28J60_ERDPTL, ad & 0xff);
    enc28j60_write_cont_reg(E28J60_ERDPTH, ad >> 8);
    currxad = ad;
  }
  spi_ss_lo();
  spi_sendbyte(0x20 | 0x1a /* "constant" */);
  res = spi_readbyte();
  spi_ss_hi();
  if (currxad == RXBUFEND) {
    currxad = 0; /* RXBUFSTART */
  } else {
    if (currxad == 0x1fff) {
      currxad = 0;
    } else {
      currxad++;
    }
  }
  return res;
}

static uint8_t enc28j60_read_rx_memory(uint16_t ad) {
  if (ad > RXBUFEND) {
    ad -= (RXBUFEND + 1);
  }
  return enc28j60_read_memory(ad);
}

static void enc28j60_write_memory(uint16_t ad, uint8_t val) {
  ad &= 0x1fff;
  if (ad != curtxad) { /* move pointer */
    enc28j60_write_cont_reg(E28J60_EWRPTL, ad & 0xff);
    enc28j60_write_cont_reg(E28J60_EWRPTH, ad >> 8);
    curtxad = ad;
  }
  spi_ss_lo();
  spi_sendbyte(0x60 | 0x1a /* "constant" */);
  spi_sendbyte(val);
  spi_ss_hi();
  if (curtxad == 0x1fff) {
    curtxad = 0;
  } else {
    curtxad++;
  }
}

static void enc28j60_write_phy(uint8_t address, uint16_t val) {
  /* set the PHY register address */
  enc28j60_write_cont_reg(E28J60_MIREGADR, address);
  /* write the PHY data */
  enc28j60_write_cont_reg(E28J60_MIWRL, val & 0xff);
  enc28j60_write_cont_reg(E28J60_MIWRH, val >> 8);
  /* wait until the PHY write completes */
  while(enc28j60_read_cont_reg(E28J60_MISTAT) & E28J60_MISTAT_BUSY) {
  }
}

static void net_filleui64(uint8_t * ad) {
  ad[0] = net_mac[0] ^ 0x02;
  ad[1] = net_mac[1];
  ad[2] = net_mac[2];
  ad[3] = 0xff;
  ad[4] = 0xfe;
  ad[5] = net_mac[3];
  ad[6] = net_mac[4];
  ad[7] = net_mac[5];
}


/* We can get an IRQ on INT1 from the NIC. This is only used to wake
 * the MCU from sleep, so empty. */
ISR(INT1_vect) {
  return;
}

/* Check if no send is in progress and there are things to send, and if
 * there are, starts a send. */
static void net_trysending()
{
  uint8_t i;
  if (txsending == 0) {
    for (i = 0; i < NUMTXBUFS; i++) {
      if (txbufused[i]) {
        txsending = i + 1;
        enc28j60_write_cont_reg(E28J60_ETXSTL, txbufstart[i] & 0xff);
        enc28j60_write_cont_reg(E28J60_ETXSTH, txbufstart[i] >> 8);
        enc28j60_write_cont_reg(E28J60_ETXNDL, (txbufstart[i] + txbufpsize[i]) & 0xff);
        enc28j60_write_cont_reg(E28J60_ETXNDH, (txbufstart[i] + txbufpsize[i]) >> 8);
        enc28j60_bitfield_set(E28J60_ECON1, E28J60_ECON1_TXRTS); /* start Send */
        // console_printpgm_P(PSTR(" txs "));
        break;
      }
    }
  }
}

/* Gets a buffer for sending a packet. Initializes it with elementary
 * stuff like own MAC. Returns the number of the buffer.
 * If no empty buffers are available, waits until there is one. */
static uint8_t net_preparetxbuffer(void) {
  uint8_t i;
  for (i = 0; i < NUMTXBUFS; i++) {
    if (txbufused[i] == 0) {
      uint8_t j;
      txbufused[i] = 1;
      txbufpsize[i] = 0;
      /* Control byte: No overrides. */
      enc28j60_write_memory(txbufstart[i], 0);
      for (j = 0; j < 6; j++) {
        enc28j60_write_memory(txbufstart[i] + 7 + j, net_mac[j]);
      }
      return i;
    }
  }
  /* Nothing free. wait for send to complete */
  if (txsending == 0) { /* Well then try to send something. */
    net_trysending();
#if 0
    console_printpgm_P(PSTR("\r\nFATAL out of buffers and not sending"));
    /* We should probably do something along the lines of cli(); hlt();
     * to force a reset by watchdog */
    return 0;
#endif
  }
  while (enc28j60_read_cont_reg(E28J60_ECON1) & E28J60_ECON1_TXRTS) {
    /* Still transmitting so wait */
  }
  txbufused[txsending - 1] = 0;
  txsending = 0;
  return net_preparetxbuffer(); /* Yes, recursive. But we know this time
                                 * a buffer will be free because we just
                                 * freed it. */
}

static uint16_t net_calcincvipchecksum(uint16_t startsum, uint16_t toadd) {
  uint32_t tmpsum = (startsum ^ 0xffffUL);
  tmpsum += toadd;
  while ((tmpsum >> 16)) {
    tmpsum = (tmpsum >> 16) + (tmpsum & 0xffffUL);
  }
  return (uint16_t)(tmpsum ^ 0xffff);
}

static uint16_t net_calcincipchecksum(uint16_t startsum, uint16_t addr, uint16_t len) {
  uint32_t tmpsum = (startsum ^ 0xffffUL);
#if 0
  console_printpgm_P(PSTR("\r\nTS"));
  console_printhex8(tmpsum >> 24);
  console_printhex8(tmpsum >> 16);
  console_printhex8(tmpsum >>  8);
  console_printhex8(tmpsum >>  0);
#endif
#if 0
  console_printpgm_P(PSTR("\r\ns"));
  console_printhex8(addr >>  8);
  console_printhex8(addr >>  0);
  console_printpgm_P(PSTR("l"));
  console_printhex8(len >>  8);
  console_printhex8(len >>  0);
  console_printhex8(enc28j60_read_memory(addr + 0));
  console_printhex8(enc28j60_read_memory(addr + 1));
#endif
  while (len > 1) {
    tmpsum += (((uint16_t)enc28j60_read_memory(addr + 0) << 8)
                       | (enc28j60_read_memory(addr + 1) << 0));
#if 0
    if (len <= 2) {
      console_printhex8(enc28j60_read_memory(addr + 0));
      console_printhex8(enc28j60_read_memory(addr + 1));
    }
#endif
    addr += 2;
    len -= 2;
  }
  if (len > 0) {
    tmpsum += (uint32_t)enc28j60_read_memory(addr + 0) << 8;
    /*console_printpgm_P(PSTR("!"));*/
  }
  while ((tmpsum >> 16)) {
    tmpsum = (tmpsum >> 16) + (tmpsum & 0xffffUL);
  }
  return (uint16_t)(tmpsum ^ 0xffff);
}

static uint16_t net_calcipchecksum(uint16_t addr, uint16_t len) {
  return net_calcincipchecksum(0xffff, addr, len);
}

/* fills IP headers by copying from a request packet given in src
 * to dst.
 * Despite the name it also copies ethernet source to ethernet destination
 * and the ethernet type field.
 * This will assume an IPv6 mode if the header length is that of IPV6,
 * although it could be IPv4 with a weird combination of options. */
static void copyfillipheaders(uint16_t dst, uint8_t iphlen) {
  uint16_t i;
  /* fill in destination */
  for (i = 0; i < 6; i++) { /* copy MAC from request */
    enc28j60_write_memory(dst + i, hdrbuf[6 + i]);
  }
  /* copy all headers up to hlen for starters */
  for (i = 12; i < iphlen; i++) {
    enc28j60_write_memory(dst + i, hdrbuf[i]);
  }
  if (iphlen == (14+40)) { /* Just assume it's IPv6 although it could be IPv4
                            * with a weird combination of options */
    /* now fix the src / dst ip */
    for (i = 0; i < 16; i++) {
      /* copy src IP from request to dst ip in reply */
      enc28j60_write_memory(dst + 38 + i, hdrbuf[22 + i]);
      /* fill in new src ip */
      enc28j60_write_memory(dst + 22 + i, net_ip6[i]);
    }
    /* Set TTL */
    enc28j60_write_memory((dst + 21), NET_TTL);
    /* There is no checksum in the IP6 header */
  } else {
    /* now fix the src / dst ip */
    for (i = 0; i < 4; i++) {
      /* copy src IP from request to dst ip in reply */
      enc28j60_write_memory(dst + 30 + i, hdrbuf[26 + i]);
      /* fill in new src ip */
      enc28j60_write_memory(dst + 26 + i, net_ip4[i]);
    }
    /* Set TTL */
    enc28j60_write_memory((dst + 22), NET_TTL);
    /* Null the IP checksum */
    enc28j60_write_memory((dst + 24), 0);
    enc28j60_write_memory((dst + 25), 0);
  }
}

static void net_handlearp() {
  /* 14 - 15  hardware type
   * 16 - 17  protocol
   * 18       hardware address size
   * 19       protocol address size
   * 20 - 21  type (0x0001 = request, 0x0002 = reply)
   * 22 - 27  sender hw address
   * 28 - 31  sender ip address
   * 32 - 37  target hw address
   * 38 - 41  target ip address
   */
  if ((hdrbuf[14] == 0x00) /* HW type eth */
   && (hdrbuf[15] == 0x01)
   && (hdrbuf[16] == 0x08) /* proto IP */
   && (hdrbuf[17] == 0x00)
   && (hdrbuf[18] == 0x06) /* HW size */
   && (hdrbuf[19] == 0x04) /* protocol size */
   && (hdrbuf[20] == 0x00) /* arp REQUEST */
   && (hdrbuf[21] == 0x01)) {
      /* OK, so this really is a proper IPv4 ARP request. */
      uint8_t i; uint16_t txad; uint8_t txbnum;
      if (memcmp(&hdrbuf[38], &net_ip4[0], 4) != 0) {
        /* Not for us */
        return;
      }
      /* the requested IP is ours. Send ARP reply. */
      txbnum = net_preparetxbuffer();
      txad = txbufstart[txbnum] + 1;
      /* fill in destination */
      for (i = 0; i < 6; i++) {
        enc28j60_write_memory(txad + i, hdrbuf[6 + i]);
      }
      /* copy the ARP "headers" from request */
      for (i = 12; i <= 20; i++) {
        enc28j60_write_memory(txad + i, hdrbuf[i]);
      }
      enc28j60_write_memory(txad + 21, 0x02); /* arp REPLY */
      /* copy sender- to target- addresses */
      for (i = 22; i <= 31; i++) {
        enc28j60_write_memory(txad + 10 + i, hdrbuf[i]);
      }
      /* Fill in new sender IP and MAC */
      for (i = 0; i < 6; i++) {
        enc28j60_write_memory(txad + 22 + i, net_mac[i]);
      }
      for (i = 0; i < 4; i++) {
        enc28j60_write_memory(txad + 28 + i, net_ip4[i]);
      }
      /* fill in length */
      txbufpsize[txbnum] = 42;
      /* and we're done. */
      return; /* redundant */
  }
}

static void net_handleicmp4(uint8_t hlen) {
  /* ICMP: IP Header + type (1 byte) + code (1 byte) + checksum (2 bytes)
   *                 + parameters (0-4 bytes) + data
   *       type  8 code 0:  echo request
   *       type  0 code 0:  echo reply
   *       type 11 code 0:  TTL exceeded
   */
  if ((hdrbuf[hlen + 0] == 8) && (hdrbuf[hlen + 1] == 0)) {
    /* We have an echo request - send it back almost unmodified */
    uint16_t cs; uint16_t txad; uint8_t txbnum; uint16_t i;
    if (plen > 511) {
      /* We don't do more than 511 bytes */
      return;
    }
    /* console_printpgm_P(PSTR(" ping?pong! ")); */
    txbnum = net_preparetxbuffer();
    txad = txbufstart[txbnum] + 1;
    copyfillipheaders(txad, hlen);
    /* copy entire rest of packet */
    for (i = hlen; i < plen; i++) {
      enc28j60_write_memory(txad + i, enc28j60_read_rx_memory(thispacketstart + i));
    }
    /* type 0 (+ code 0 copied over): echo reply */
    enc28j60_write_memory((txad + hlen + 0), 0);
    /* Calculate checksums. As this calculation includes the checksum itself,
     * we must set it to 0 before the calculation. */
    /* ICMP checksum */
    enc28j60_write_memory((txad + hlen + 2), 0);
    enc28j60_write_memory((txad + hlen + 3), 0);
    cs = net_calcipchecksum((txad + hlen), plen - hlen);
    enc28j60_write_memory((txad + hlen + 2), (cs >> 8));
    enc28j60_write_memory((txad + hlen + 3), (cs & 0xff));
    /* IP checksum */
    cs = net_calcipchecksum(txad + 14, hlen - 14);
    enc28j60_write_memory((txad + 24), (cs >> 8));
    enc28j60_write_memory((txad + 25), (cs & 0xff));
    txbufpsize[txbnum] = plen;
    /* Packet ready to be sent */
    return; /* redundant */
  }
}

/* Returns: Number of bytes in reply packet. 0 if no reply should be sent. */
static uint16_t net_handlentp(uint16_t srcstart, uint16_t dststart, uint16_t len) {
  struct timestamp dts; /* device timestamp */
  uint32_t tsdif;
  uint32_t tstmp;
  uint16_t i;
  uint8_t mygpsv = gpsvalid;
  if (len < 48) {
    ERRORconsole_pgm_P(PSTR(" NTP-invlen "));
    return 0;
  }
  if (gpstime[mygpsv].unixts == 0) { /* We don't have a valid time */
    return 0;
  }
  enc28j60_write_memory((dststart +  0), 0x1c); /* Flags: No Leap, V3, Server */
  enc28j60_write_memory((dststart +  1), 0x01); /* Stratum */
  enc28j60_write_memory((dststart +  2), 0x0a); /* Poll interval = 2^10 = 1024 sec */
  enc28j60_write_memory((dststart +  3), 0xfa); /* Precision 2^-6 s */
  enc28j60_write_memory((dststart +  4), 0x00); /* Some fake small root delay */
  enc28j60_write_memory((dststart +  5), 0x00); /* Some fake small root delay */
  enc28j60_write_memory((dststart +  6), 0x00); /* Some fake small root delay */
  enc28j60_write_memory((dststart +  7), 0x04); /* Some fake small root delay */
  enc28j60_write_memory((dststart +  8), 0x00); /* Some fake small root dispersion */
  enc28j60_write_memory((dststart +  9), 0x00); /* Some fake small root dispersion */
  enc28j60_write_memory((dststart + 10), 0x00); /* Some fake small root dispersion */
  enc28j60_write_memory((dststart + 11), 0x08); /* Some fake small root dispersion */
  enc28j60_write_memory((dststart + 12), 'G'); /* reference identifier */
  enc28j60_write_memory((dststart + 13), 'P'); /* reference identifier */
  enc28j60_write_memory((dststart + 14), 'S'); /* reference identifier */
  enc28j60_write_memory((dststart + 15),   0); /* reference identifier */
  /* fill in time info */
  /* Reference Timestamp */
  tstmp = gpstime[mygpsv].unixts + 2208988800UL; /* Yes, this WILL overflow in 2036 */
  enc28j60_write_memory((dststart + 16), (uint8_t)(tstmp >> 24));
  enc28j60_write_memory((dststart + 17), (uint8_t)(tstmp >> 16));
  enc28j60_write_memory((dststart + 18), (uint8_t)(tstmp >>  8));
  enc28j60_write_memory((dststart + 19), (uint8_t)(tstmp >>  0));
  enc28j60_write_memory((dststart + 20), 0);
  enc28j60_write_memory((dststart + 21), 0);
  enc28j60_write_memory((dststart + 22), 0);
  enc28j60_write_memory((dststart + 23), 0);
  /* Originate Timestamp */
  for (i = 0; i < 8; i++) {
    enc28j60_write_memory((dststart + 24 + i),
              enc28j60_read_rx_memory(thispacketstart + srcstart + 40 + i));
  }
  gettimestamp(&dts);
  tsdif = ((uint32_t)dts.tovfls << 16) | dts.ticks;
  tsdif -= ((uint32_t)gpstime[mygpsv].devicetime.tovfls << 16)
           | gpstime[mygpsv].devicetime.ticks;
  tsdif <<= 16;
  tsdif += dts.partticks - gpstime[mygpsv].devicetime.partticks;
  tstmp += (tsdif / CPUFREQ);
  /* Receive and Transmit TS (whole seconds) */
  enc28j60_write_memory((dststart + 32), (uint8_t)(tstmp >> 24));
  enc28j60_write_memory((dststart + 33), (uint8_t)(tstmp >> 16));
  enc28j60_write_memory((dststart + 34), (uint8_t)(tstmp >>  8));
  enc28j60_write_memory((dststart + 35), (uint8_t)(tstmp >>  0));
  enc28j60_write_memory((dststart + 40), (uint8_t)(tstmp >> 24));
  enc28j60_write_memory((dststart + 41), (uint8_t)(tstmp >> 16));
  enc28j60_write_memory((dststart + 42), (uint8_t)(tstmp >>  8));
  enc28j60_write_memory((dststart + 43), (uint8_t)(tstmp >>  0));
  /* Receive and Transmit TS (fraction) */
  tstmp = (0xffffffffUL / CPUFREQ) * (tsdif % CPUFREQ);
  enc28j60_write_memory((dststart + 36), (uint8_t)(tstmp >> 24));
  enc28j60_write_memory((dststart + 37), (uint8_t)(tstmp >> 16));
  enc28j60_write_memory((dststart + 38), (uint8_t)(tstmp >>  8));
  enc28j60_write_memory((dststart + 39), (uint8_t)(tstmp >>  0));
  enc28j60_write_memory((dststart + 44), (uint8_t)(tstmp >> 24));
  enc28j60_write_memory((dststart + 45), (uint8_t)(tstmp >> 16));
  enc28j60_write_memory((dststart + 46), (uint8_t)(tstmp >>  8));
  enc28j60_write_memory((dststart + 47), (uint8_t)(tstmp >>  0));
  return 48;
}

/* Returns: Number of bytes in reply packet. 0 if no reply should be sent. */
static uint16_t net_handlercon(uint16_t srcstart, uint16_t dststart, uint16_t len) {
  /* Note: at the default HEADERBUFSIZE of 84 we can go up to
   * hdrbuf[srcstart + 21] */
  uint16_t i; uint8_t cmd;
  if (len < 21) { /* minimum length for ANY command */
    return 0;
  }
  if ((hdrbuf[srcstart + 0] != RCON_MAGIC1)
   || (hdrbuf[srcstart + 1] != RCON_MAGIC2)) { /* no magic marker */
    return 0;
  }
  for (i = 0; i < 16; i++) {
    if (hdrbuf[srcstart + 2 + i] != net_rconpass[i]) { /* Password incorrect */
      ERRORconsole_pgm_P(PSTR("  badrcpw  "));
      return 0;
    }
  }
  enc28j60_write_memory((dststart +  0), RCON_MAGIC1);
  enc28j60_write_memory((dststart +  1), RCON_MAGIC2);
  enc28j60_write_memory((dststart +  2), hdrbuf[srcstart + 18]);
  enc28j60_write_memory((dststart +  3), hdrbuf[srcstart + 19]);
  cmd = hdrbuf[srcstart + 20];
  if (cmd == RCON_CMD_PING) {
    enc28j60_write_memory((dststart +  4), RCON_REP_ACK);
    return 5;
  } else if (cmd == RCON_CMD_SETBRI) {
    uint8_t bri = hdrbuf[srcstart + 21];
    if (bri < 8) {
      led_brightness = bri;
      enc28j60_write_memory((dststart +  4), RCON_REP_ACK);
    } else {
      enc28j60_write_memory((dststart +  4), RCON_REP_NAK);
    }
    enc28j60_write_memory((dststart +  5), led_brightness);
    return 6;
  } else if (cmd == RCON_CMD_SETDISPOV) {
    if ((len < 28) || (hdrbuf[srcstart + 21] > 50)) {
      enc28j60_write_memory((dststart +  4), RCON_REP_NAK);
      return 5;
    }
    led_dispovon = 1;
    led_dispovendticks = ((hdrbuf[srcstart + 21] * TICKSPERSECOND) / 10);
    led_dispovendticks += getticks();
    for (i = 0; i < 6; i++) {
      led_dispovbytes[i] = enc28j60_read_rx_memory(thispacketstart + srcstart + 22 + i);
    }
    enc28j60_write_memory((dststart +  4), RCON_REP_ACK);
    return 5;
  }
  /* If we reach this we do not know the command. */
  enc28j60_write_memory((dststart +  4), RCON_REP_NAK);
  return 5;
}

/* for reference:
 * UDP: IP Header + src port (2 bytes) + dst port (2 bytes)
 *                + length (2 bytes) + checksum (2 bytes, optional on IPv4)
 *                + data
 */
static void net_handleudp4(uint8_t hlen) {
  if (((hdrbuf[hlen + 2] ==   0) && (hdrbuf[hlen + 3] == 123))    /* dst Port 123 (NTP) */
   || ((hdrbuf[hlen + 2] == 122) && (hdrbuf[hlen + 3] == 105))) { /* dst port 31337 (rcon) */
    /* We listen on this port */
    uint16_t cs; uint16_t txad; uint8_t txbnum; uint16_t i;
    if (plen > 511) { /* check for remotely sane length */
      return;
    }
    txbnum = net_preparetxbuffer();
    txad = txbufstart[txbnum] + 1;
    copyfillipheaders(txad, hlen);
    if ((hdrbuf[hlen + 2] ==   0) && (hdrbuf[hlen + 3] == 123)) { /* NTP */
      i = net_handlentp(hlen+8, txad+hlen+8, plen-hlen-8);
    } else { /* rcon */
      i = net_handlercon(hlen+8, txad+hlen+8, plen-hlen-8);
    }
    if (i == 0) { /* the NTP / RCON handler does not want to send a reply */
      txbufused[txbnum] = 0;
      return;
    }
    /* copy port numbers in UDP header */
    enc28j60_write_memory((txad + hlen + 0), hdrbuf[hlen + 2]);
    enc28j60_write_memory((txad + hlen + 1), hdrbuf[hlen + 3]);
    enc28j60_write_memory((txad + hlen + 2), hdrbuf[hlen + 0]);
    enc28j60_write_memory((txad + hlen + 3), hdrbuf[hlen + 1]);
    /* fill in length of UDP, we abuse cs as temporary */
    cs = i + 8;
    enc28j60_write_memory((txad + hlen + 4), (cs >> 8));
    enc28j60_write_memory((txad + hlen + 5), (cs & 0xff));
    /* fill in length of IP packet */
    cs = i + 20 + 8;
    enc28j60_write_memory((txad + 16), (cs >> 8));
    enc28j60_write_memory((txad + 17), (cs & 0xff));
    /* Calculate checksums. As this calculation includes the checksum itself,
     * we must set it to 0 before the calculation. */
    /* UDP checksum. This is great fun because of the FSCKing "pseudo header" */
    enc28j60_write_memory((txad + hlen + 6), 0);
    enc28j60_write_memory((txad + hlen + 7), 0);
    cs = net_calcipchecksum((txad + hlen), i + 8);
    /* Now the IPs in the pseudo header. Fun fun fun. */
    cs = net_calcincipchecksum(cs, (txad + 26), 8); /* src + dst IP */
    cs = net_calcincvipchecksum(cs, 0x0011); /* zeros and protocol number */
    cs = net_calcincvipchecksum(cs, i + 8); /* UDP length */
    enc28j60_write_memory((txad + hlen + 6), (cs >> 8));
    enc28j60_write_memory((txad + hlen + 7), (cs & 0xff));
    /* IP checksum */
    cs = net_calcipchecksum(txad + 14, hlen - 14);
    enc28j60_write_memory((txad + 24), (cs >> 8));
    enc28j60_write_memory((txad + 25), (cs & 0xff));
    txbufpsize[txbnum] = hlen + 8 + i;
    /* Packet ready to be sent */
    return; /* redundant */
  }
}

/* for reference:
 * UDP: IP Header + src port (2 bytes) + dst port (2 bytes)
 *                + length (2 bytes) + checksum (2 bytes)
 *                + data
 */
static void net_handleudp6(void) {
  uint16_t hlen = 14 + 40; /* The header length is fixed on IPv6 */
  if (((hdrbuf[hlen + 2] ==   0) && (hdrbuf[hlen + 3] == 123))    /* dst Port 123 (NTP) */
   || ((hdrbuf[hlen + 2] == 122) && (hdrbuf[hlen + 3] == 105))) { /* dst port 31337 (rcon) */
    /* We listen on this port */
    uint16_t cs; uint16_t txad; uint8_t txbnum; uint16_t i;
    if (plen > 511) { /* check for remotely sane length */
      return;
    }
    txbnum = net_preparetxbuffer();
    txad = txbufstart[txbnum] + 1;
    copyfillipheaders(txad, hlen);
    if ((hdrbuf[hlen + 2] ==   0) && (hdrbuf[hlen + 3] == 123)) { /* NTP */
      i = net_handlentp(hlen+8, txad+hlen+8, plen-hlen-8);
    } else { /* rcon */
      i = net_handlercon(hlen+8, txad+hlen+8, plen-hlen-8);
    }
    if (i == 0) { /* the NTP / RCON handler does not want to send a reply */
      txbufused[txbnum] = 0;
      return;
    }
    /* copy port numbers in UDP header */
    enc28j60_write_memory((txad + hlen + 0), hdrbuf[hlen + 2]);
    enc28j60_write_memory((txad + hlen + 1), hdrbuf[hlen + 3]);
    enc28j60_write_memory((txad + hlen + 2), hdrbuf[hlen + 0]);
    enc28j60_write_memory((txad + hlen + 3), hdrbuf[hlen + 1]);
    /* fill in length of UDP, we abuse cs as temporary */
    cs = i + 8;
    enc28j60_write_memory((txad + hlen + 4), (cs >> 8));
    enc28j60_write_memory((txad + hlen + 5), (cs & 0xff));
    /* fill in length of IP packet - equal to length in UDP header on V6! */
    enc28j60_write_memory((txad + 18), (cs >> 8));
    enc28j60_write_memory((txad + 19), (cs & 0xff));
    /* Calculate checksums. As this calculation includes the checksum itself,
     * we must set it to 0 before the calculation. */
    /* UDP checksum. This is great fun because of the FSCKing "pseudo header" */
    enc28j60_write_memory((txad + hlen + 6), 0);
    enc28j60_write_memory((txad + hlen + 7), 0);
    cs = net_calcipchecksum((txad + hlen), i + 8);
    /* Now the IPs in the pseudo header. Fun fun fun. */
    cs = net_calcincipchecksum(cs, (txad + 22), 32); /* src + dst IP */
    cs = net_calcincvipchecksum(cs, 0); /* Zeros (no kidding!) */
    cs = net_calcincvipchecksum(cs, 0x0011); /* Zeros and protocol number */
    cs = net_calcincvipchecksum(cs, 0);     /* The UDP length (upper 16 bit) */
    cs = net_calcincvipchecksum(cs, i + 8); /* The UDP length (lower 16 bit) */
    enc28j60_write_memory((txad + hlen + 6), (cs >> 8));
    enc28j60_write_memory((txad + hlen + 7), (cs & 0xff));
    txbufpsize[txbnum] = hlen + 8 + i;
    /* Packet ready to be sent */
    return; /* redundant */
  }
}

static void net_handletcp4(uint8_t iphlen) {
  /* TCP: IP Header +
   * Byte 0+1: srcport, 2+3 dstport
   * Byte 4-7: sequence number
   * Byte 8-11: acknowledgement number
   * Byte 12+13: various.
   *   Bit 0-3: data offset (in multiples of 4 bytes) = number of bytes in TCP header
   *   Bit 4-6: reserved (0)
   *   Bit 7-9: ECN, Bit 10: URG, Bit 11: ACK, Bit 12: PSH, Bit 13: RST,
   *   Bit 14: SYN, Bit 15: FIN
   * Byte 14+15: window
   * Byte 16+17: checksum
   * Byte 18+19: urgent pointer
   */
  uint16_t txad;
  uint8_t txbnum;
  uint16_t i;
  uint16_t sport;
  uint16_t dport;
  uint16_t tcphlen;
  uint32_t seqnr;
  
  sport = (((uint16_t)enc28j60_read_rx_memory(thispacketstart + iphlen + 0) << 8)
        | enc28j60_read_rx_memory(thispacketstart + iphlen + 1) << 0);
  dport = (((uint16_t)enc28j60_read_rx_memory(thispacketstart + iphlen + 2) << 8)
        | enc28j60_read_rx_memory(thispacketstart + iphlen + 3) << 0);
  tcphlen = (enc28j60_read_rx_memory(thispacketstart + iphlen + 12) >> 4) * 4;
  /* Insert sanity checks here */
  /* prepare reply (either RST for closed ports, or a real reply */
  txbnum = net_preparetxbuffer();
  txad = txbufstart[txbnum] + 1;
  /* Fill IP headers from request packet */
  copyfillipheaders(txad, iphlen);
  /* insert ports (dst + src swapped) */
  enc28j60_write_memory(txad + iphlen + 0, dport >> 8);
  enc28j60_write_memory(txad + iphlen + 1, dport & 0xff);
  enc28j60_write_memory(txad + iphlen + 2, sport >> 8);
  enc28j60_write_memory(txad + iphlen + 3, sport & 0xff);
  seqnr = (uint32_t)enc28j60_read_rx_memory(thispacketstart + iphlen + 4 + 0) << 24
        | (uint32_t)enc28j60_read_rx_memory(thispacketstart + iphlen + 4 + 1) << 16
        | (uint32_t)enc28j60_read_rx_memory(thispacketstart + iphlen + 4 + 2) <<  8
        | (uint32_t)enc28j60_read_rx_memory(thispacketstart + iphlen + 4 + 3) <<  0;
  for (i = 0; i < 4; i++) {
    /* clear checksum and urgent pointer */
    enc28j60_write_memory(txad + iphlen + 16 + i, 0);
  }
  /* Header size: always 5*4 = 20 bytes, we don't do options. */
  enc28j60_write_memory(txad + iphlen + 12, 5 << 4);
  /* set length in IP header */
  i = iphlen - 14 + 20; /* - ethernet + tcp header size */
  enc28j60_write_memory(txad + 14 + 2, i >> 8);
  enc28j60_write_memory(txad + 14 + 3, i & 0xff);
  /* Fill IP checksum (abusing i as variable for checksum) */
  i = net_calcipchecksum(txad + 14, iphlen - 14);
  enc28j60_write_memory((txad + 24), (i >> 8));
  enc28j60_write_memory((txad + 25), (i & 0xff));
  
  if (1) {
    /* default handling: unknown port, send TCP reset.
     * This gets overwritten below _if_ the packet gets properly handled. */
    for (i = 0; i < 4; i++) {
      /* null sequence number */
      enc28j60_write_memory(txad + iphlen + 4 + i, 0);
    }
    /* copy original sequence to ack number - and increase by 1. */
    enc28j60_write_memory(txad + iphlen + 8 + 0, (seqnr + 1) >> 24);
    enc28j60_write_memory(txad + iphlen + 8 + 1, (seqnr + 1) >> 16);
    enc28j60_write_memory(txad + iphlen + 8 + 2, (seqnr + 1) >>  8);
    enc28j60_write_memory(txad + iphlen + 8 + 3, (seqnr + 1) >>  0);
    /* Set window to zero */
    enc28j60_write_memory(txad + iphlen + 14, 0);
    enc28j60_write_memory(txad + iphlen + 15, 0);
    /* set SYN and RST */
    enc28j60_write_memory(txad + iphlen + 13, TCP_SYN | TCP_RST);
  }
  
#ifdef NETCONSOLE
  if (dport == TELNETPORT) {
    uint8_t tcpflags = enc28j60_read_rx_memory(thispacketstart + iphlen + 13);
    /* Accept / handle connection */
    for (i = 0; i < NUMTCPCONNS; i++) {
      if ((tcpconns[i].iptype == 0) /* IPv4 connection */
       && (tcpconns[i].myport == TELNETPORT) /* to this port */
       && (tcpconns[i].partnerport == sport)) { /* from this port */
        uint8_t j;
        for (j = 0; j < 4; j++) {
          if (enc28j60_read_rx_memory(thispacketstart + 26 + j) != tcpconns[i].partnerip4[j]) {
            break;
          }
        }
        if (j == 4) { /* Success */
          break;
        }
      }
    }
    if (i < NUMTCPCONNS) { /* We found our slot (existing connection) */
      uint32_t acknr = (uint32_t)enc28j60_read_rx_memory(thispacketstart + iphlen + 8 + 0) << 24
                     | (uint32_t)enc28j60_read_rx_memory(thispacketstart + iphlen + 8 + 1) << 16
                     | (uint32_t)enc28j60_read_rx_memory(thispacketstart + iphlen + 8 + 2) <<  8
                     | (uint32_t)enc28j60_read_rx_memory(thispacketstart + iphlen + 8 + 3) <<  0;
      if (seqnr == tcpconns[i].partnersnr) {
        if (acknr == tcpconns[i].mysnr) {
          uint16_t payloadlen;
          uint8_t c;
          uint8_t pstate = 0;
          
          payloadlen = (uint16_t)enc28j60_read_rx_memory(thispacketstart + 14 + 2) << 8
                     | enc28j60_read_rx_memory(thispacketstart + 14 + 3) << 0;
          payloadlen = payloadlen - tcphlen - (iphlen - 14);
//           console_printpgm_P(PSTR(" OC "));
          /* OK, this really is our connection. */
          if (tcpconns[i].state == STATE_SYNACKSENT) {
            /* Only ACK should be set */
            if ((tcpflags & TCP_RELEVANTFLAGMASK) == TCP_ACK) {
//               console_printpgm_P(PSTR(" CONEST "));
              tcpconns[i].state = STATE_ESTABLISHED;
            }
          }
          if (tcpconns[i].state == STATE_LASTACK) {
            /* This will be the last ACK we send. Mark the connection
             * state entry as free. */
            tcpconns[i].myport = 0;
          }
          /* Our own sequence number. */
          /* do not increase, we did not send anything. */
          enc28j60_write_memory(txad + iphlen + 4 + 0, tcpconns[i].mysnr >> 24);
          enc28j60_write_memory(txad + iphlen + 4 + 1, tcpconns[i].mysnr >> 16);
          enc28j60_write_memory(txad + iphlen + 4 + 2, tcpconns[i].mysnr >>  8);
          enc28j60_write_memory(txad + iphlen + 4 + 3, tcpconns[i].mysnr >>  0);
          /* By default we only ACKnowledge */
          enc28j60_write_memory(txad + iphlen + 13, TCP_ACK);
          if (tcpflags & TCP_FIN) {
            /* OK, so this connection closes. */
            /* Send back a FIN / ACK. */
            enc28j60_write_memory(txad + iphlen + 13, TCP_ACK | TCP_FIN);
            /* Mark it as last ack so we send back a last acknowledgement */
            tcpconns[i].state = STATE_LASTACK;
            /* It is on purpose that we increase mysnr AFTER we put it into
             * the packet, and partnersnr BEFORE. */
            tcpconns[i].partnersnr++;
            tcpconns[i].mysnr++;
          } else { /* Not closing yet */
            tcpconns[i].lastactive = getticks();
          }
          /* Set window */
          enc28j60_write_memory(txad + iphlen + 14, 1500 >> 8);
          enc28j60_write_memory(txad + iphlen + 15, 1500 & 0xff);
          /* Acknowledge payload */
          tcpconns[i].partnersnr += payloadlen;
          enc28j60_write_memory(txad + iphlen + 8 + 0, tcpconns[i].partnersnr >> 24);
          enc28j60_write_memory(txad + iphlen + 8 + 1, tcpconns[i].partnersnr >> 16);
          enc28j60_write_memory(txad + iphlen + 8 + 2, tcpconns[i].partnersnr >>  8);
          enc28j60_write_memory(txad + iphlen + 8 + 3, tcpconns[i].partnersnr >>  0);
          /*console_printpgm_P(PSTR("l"));
          console_printhex8(payloadlen >> 8);
          console_printhex8(payloadlen >> 0);*/
          /* Now handle the payload. WARNING OVERWRITES i YOU CANNOT USE
           * tcpconns[i] AFTERWARDS */
          for (i = 0; i < payloadlen; i++) {
            c = enc28j60_read_rx_memory(thispacketstart + iphlen + tcphlen + i);
            if (pstate == 0) { /* No "IAC" crap */
              if (c == 255) { /* Special telnet command follows */
                pstate = 1;
              } else { /* Normal character, feed into input buffer */
                cli();
                console_inputchar(c);
                sei();
              }
            } else if (pstate == 1) {
              /* First character after escape */
              switch (c) {
              case 250: /* Start of option negotiation. Read until char 240 */
                           pstate = 3; break;
              case 251 ... 254: /* Gets one parameter to ignore */
                           pstate = 2; break;
              default: /* no idea. just go back to normal. */
                           pstate = 0; break;
              };
            } else if (pstate == 2) {
              /* Second character after escape */
              pstate = 0;
            } else if (pstate == 3) {
              if (c == 240) { /* End of option negotiation. */
                pstate = 0;
              }
            }
          }
        } else {
          uint32_t dist = acknr - tcpconns[i].mysnr;
          if (dist < 5000) { /* Within a reasonable range */
            /* Do not send RST in this case */
            txbufused[txbnum] = 0;
          }
        }
      } else {
        uint32_t dist = acknr - tcpconns[i].mysnr;
        dist += seqnr;
        dist -= tcpconns[i].partnersnr;
        if (dist < 10000) { /* Within a reasonable range */
          /* Do not send RST in this case */
          txbufused[txbnum] = 0;
        }
      }
    } else { /* New connection? */
      /* Only SYN should be set */
      if ((tcpflags & TCP_RELEVANTFLAGMASK) == TCP_SYN) {
//         console_printpgm_P(PSTR(" NC "));
        for (i = 0; i < NUMTCPCONNS; i++) {
          if (tcpconns[i].myport == 0) { /* unused slot. break. */
            break;
          }
        }
        if (i < NUMTCPCONNS) { /* Usable slot */
          uint8_t j;
          /* save state */
          tcpconns[i].iptype = 0;
          tcpconns[i].myport = dport;
          tcpconns[i].partnerport = sport;
          for (j = 0; j < 4; j++) {
            tcpconns[i].partnerip4[j] = enc28j60_read_rx_memory(thispacketstart + 26 + j);
          }
          for (j = 0; j < 6; j++) {
            tcpconns[i].partnermac[j] = enc28j60_read_rx_memory(thispacketstart + 6 + j);
          }
          tcpconns[i].partnersnr = seqnr + 1;
          tcpconns[i].mysnr = 0xaffebeef; /* FIXME */
          tcpconns[i].state = STATE_SYNACKSENT;
          tcpconns[i].lastactive = getticks();
          /* set sequence number */
          enc28j60_write_memory(txad + iphlen + 4 + 0, tcpconns[i].mysnr >> 24);
          enc28j60_write_memory(txad + iphlen + 4 + 1, tcpconns[i].mysnr >> 16);
          enc28j60_write_memory(txad + iphlen + 4 + 2, tcpconns[i].mysnr >>  8);
          enc28j60_write_memory(txad + iphlen + 4 + 3, tcpconns[i].mysnr >>  0);
          tcpconns[i].mysnr++;
          enc28j60_write_memory(txad + iphlen + 13, TCP_SYN | TCP_ACK);
          /* Set window */
          enc28j60_write_memory(txad + iphlen + 14, 1500 >> 8);
          enc28j60_write_memory(txad + iphlen + 15, 1500 & 0xff);
        }
      }
    }
  }
#endif /* NETCONSOLE */
  /* Set length for the packet to send */
  txbufpsize[txbnum] = iphlen + 20; /* just the header, no data */
  /* add the "pseudo header" for checksum calculation:
   * Byte 0-3: srcip, Byte 4-7: dstip, Byte 8: 0, Byte 9: prot.number,
   * Byte 10+11: (TCP) length
   */
  /* We just put the pseudo fields behind the real packet for the
   * calculation. */
  for (i = 0; i < 8; i++) {
    enc28j60_write_memory(txad + iphlen + 20 + i,
                          enc28j60_read_memory(txad + 14 + 12 + i));
  }
  enc28j60_write_memory(txad + iphlen + 20 + 8, 0); /* just 0. */
  enc28j60_write_memory(txad + iphlen + 20 + 9, 0x06); /* Prot. number TCP */
  enc28j60_write_memory(txad + iphlen + 20 + 10, 20 >> 8); /* length */
  enc28j60_write_memory(txad + iphlen + 20 + 11, 20 & 0xff); /* length */
  /* TCP checksum */
  i = net_calcipchecksum(txad + iphlen, 20 + 12);
  enc28j60_write_memory((txad + iphlen + 16), (i >> 8));
  enc28j60_write_memory((txad + iphlen + 17), (i & 0xff));
  /* Packet ready to be sent! */
}

/* for reference: IP header
 * Ethernet header +
 * Byte 0: Version (4 bits) + IP Header length in mult. of 4 bytes (4 bits)
 * Byte 1: differentiated service / TOS
 * Byte 2+3: length
 * Byte 4+5: Identification for fragments
 * Byte 6+7: fragmentation flags (3 bits) + fragment offset (13 bits)
 * Byte 8: TTL
 * Byte 9: Protocol
 * Byte 10+11: Header checksum
 * Byte 12-15: Source
 * Byte 16-19: Destination
 * Byte 20...: optional options
 */
static void net_handleip4(void) {
  uint8_t hlen; /* Header length (Points to start of actual data) */
  
  if (memcmp(&hdrbuf[30], &net_ip4[0], 4) != 0) {
    /* Not for us */
    return;
  }
  hlen = hdrbuf[14];
  if ((hlen & 0xf0) != 0x40) {
    /* Not IPv_4_ */
    return;
  }
  /* hlen is given in multiples of 32 bit, thus multiply with 4 and add the
   * size of the ethernet headers to get the start of ip data */
  hlen = ((hlen & 0x0f) << 2) + 14;
  if ((hlen > plen) || (hlen > (HEADERBUFSIZE - 20))) {
    /* points behind end of packet? nonsense. */
    ERRORconsole_pgm_P(PSTR(" inv iplen "));
    return;
  }
  if (((hdrbuf[20] & 0xBF) != 0x00)
    || (hdrbuf[21] != 0x00)) {
    /* We cannot handle fragmentation. */
    ERRORconsole_pgm_P(PSTR(" -EFRAG "));
    return;
  }
  /* That should be enough sanity checks for now */
  if (hdrbuf[23] == 1) { /* Protocol ICMP */
    net_handleicmp4(hlen);
  } else if (hdrbuf[23] == 17) { /* Protocol UDP */
    net_handleudp4(hlen);
  } else if (hdrbuf[23] == 6) { /* Protocol TCP */
    net_handletcp4(hlen);
  }
}

/* For reference: after the ICMPv6 header (58 B), what follows is 4 reserved
 * bytes (0), then the requested address. Followed by Options. */
static void net_handlev6neighboursol(uint16_t v6pllen) {
  uint8_t i; uint16_t txad; uint8_t txbnum; uint16_t cs;
  
  /* Lets check if this is for us */
  if ((memcmp(&hdrbuf[62], &net_ip6[0], 16) != 0)
   && (memcmp(&hdrbuf[62], &net_llip6[0], 16) != 0)) {
    /* not requesting one of our unicast addresses. */
    return;
  }
  if (hdrbuf[21] != 255) { /* These requests MUST have a TTL of 255 */
    return;
  }
  /* This really requests our address, so create a reply */
  txbnum = net_preparetxbuffer();
  txad = txbufstart[txbnum] + 1;
  /* fill in destination */
  for (i = 0; i < 6; i++) {
    enc28j60_write_memory(txad + i, hdrbuf[6 + i]);
  }
  /* Copy original packet parts */
  for (i = 12; i < 21; i++) {
    enc28j60_write_memory(txad + i, hdrbuf[i]);
  }
  /* Set the TTL - to 255, not a real TTL! This is to make clear the
   * packet has never left the LAN! */
  enc28j60_write_memory(txad + 21, 255);
  /* and fix the payload length */
  enc28j60_write_memory(txad + 18, 0);
  enc28j60_write_memory(txad + 19, (86 - 54));
  /* Set source address to the requested address */
  for (i = 0; i < 16; i++) {
    enc28j60_write_memory(txad + 22 + i, hdrbuf[62 + i]);
  }
  /* Set destination address to the requestors address */
  for (i = 0; i < 16; i++) {
    if ((hdrbuf[22] == 0) && (hdrbuf[23] == 0)) { /* Unspecified address */
      /* (0::0) requested this (probably to check if the address is free).
       * We should check all bytes if they're really 0, but I'm too lazy.
       * Send reply to the all-nodes multicast address (although it is
       * still unicast in the ethernet header...). */
      enc28j60_write_memory(txad + 38 + i, pgm_read_byte(&ipv6_allnodes[i]));
    } else {
      enc28j60_write_memory(txad + 38 + i, hdrbuf[22 + i]);
    }
  }
  /* Now the ICMPv6 header */
  enc28j60_write_memory(txad + 54, 136); /* Type = neighbour advertisement */
  /* code, checksum and reserved bytes -> 0 */
  for (i = 55; i < 62; i++) {
    enc28j60_write_memory(txad + i, 0);
  }
  /* Fix up the flags (first 3 bits at 58):
   * Router = 0, Reply to Sol. = 1, overwrite cached value = 1 */
  enc28j60_write_memory(txad + 58, 0x60);
  /* set Target address to the requested address */
  for (i = 62; i < 78; i++) {
    enc28j60_write_memory(txad + i, hdrbuf[i]);
  }
  /* add ICMP option: Target link layer address (that's the whole point
   * of this message after all...) */
  enc28j60_write_memory(txad + 78, 0x02); /* Option type: target ll address */
  enc28j60_write_memory(txad + 79, 0x01); /* Option length (* 8 bytes) incl. header! */
  for (i = 0; i < 6; i++) {
    enc28j60_write_memory(txad + 80 + i, net_mac[i]);
  }
  /* The next 8 bytes are _not_ part of the actual message anymore, but
   * needed for the pseudo header crap in the checksum */
  for (i = 86; i < 93; i++) {
    enc28j60_write_memory(txad + i, 0);
  }
  enc28j60_write_memory(txad + 89, 32); /* length */
  enc28j60_write_memory(txad + 93, 0x3a); /* Next-header */
  /* Now calculate the checksum */
  cs = net_calcipchecksum(       txad + 22, 16);
  cs = net_calcincipchecksum(cs, txad + 38, 16);
  cs = net_calcincipchecksum(cs, txad + 86, 8);
  cs = net_calcincipchecksum(cs, txad + 54, 86 - 54);
  /* and fill in the checksum */
  enc28j60_write_memory(txad + 56, cs >> 8);
  enc28j60_write_memory(txad + 57, cs & 0xff);
  /* Ready to send */
  txbufpsize[txbnum] = 86;
}

static void net_handlev6ping(uint16_t v6pllen) {
  uint16_t i; uint16_t txad; uint8_t txbnum; uint16_t cs;
  
  /* Lets check if this is for us */
  if ((memcmp(&hdrbuf[38], &net_ip6[0], 16) != 0)
   && (memcmp(&hdrbuf[38], &net_llip6[0], 16) != 0)) {
    /* not requesting one of our unicast addresses. */
    return;
  }
  if ((hdrbuf[22] == 0) && (hdrbuf[23] == 0)) { /* Unspecified address */
    /* We do not reply to that */
    return;
  }
  if (v6pllen > 512) { /* We don't do bigger than 512 byte */
    return;
  }
  /* prepare reply */
  txbnum = net_preparetxbuffer();
  txad = txbufstart[txbnum] + 1;
  /* fill in destination */
  for (i = 0; i < 6; i++) {
    enc28j60_write_memory(txad + i, hdrbuf[6 + i]);
  }
  /* Copy original packet */
  for (i = 12; i < 21; i++) {
    enc28j60_write_memory(txad + i, hdrbuf[i]);
  }
  /* Set the TTL */
  enc28j60_write_memory(txad + 21, NET_TTL);
  /* Set source address to the requested address */
  for (i = 0; i < 16; i++) {
    enc28j60_write_memory(txad + 22 + i, hdrbuf[38 + i]);
  }
  /* Set destination address to the requestors address */
  for (i = 0; i < 16; i++) {
    enc28j60_write_memory(txad + 38 + i, hdrbuf[22 + i]);
  }
  /* Set the ICMP type: 129 (reply) instead of 128 (request) */
  enc28j60_write_memory(txad + 54, 129);
  /* Copy the rest of the packet again */
  for (i = 1; i < v6pllen; i++) {
    enc28j60_write_memory(txad + 54 + i,
                          enc28j60_read_rx_memory(thispacketstart + 54 + i));
  }
  /* Null the checksum */
  enc28j60_write_memory(txad + 56, 0);
  enc28j60_write_memory(txad + 57, 0);
  /* The next 8 bytes are _not_ part of the actual message anymore, but
   * needed for the pseudo header crap in the checksum */
  for (i = 0; i < 7; i++) {
    enc28j60_write_memory(txad + 54 + v6pllen + i, 0);
  }
  enc28j60_write_memory(txad + 54 + v6pllen + 2, v6pllen >> 8); /* length */
  enc28j60_write_memory(txad + 54 + v6pllen + 3, v6pllen >> 0); /* length */
  enc28j60_write_memory(txad + 54 + v6pllen + 7, 0x3a); /* Next-header */
  /* Now calculate the checksum */
  cs = net_calcipchecksum(       txad + 22, 16);
  cs = net_calcincipchecksum(cs, txad + 38, 16);
  cs = net_calcincipchecksum(cs, txad + 54 + v6pllen, 8);
  cs = net_calcincipchecksum(cs, txad + 54, v6pllen);
  /* and fill in the checksum */
  enc28j60_write_memory(txad + 56, cs >> 8);
  enc28j60_write_memory(txad + 57, cs & 0xff);
  /* Ready to send */
  txbufpsize[txbnum] = 54 + v6pllen;
}

/* For reference: Route Adv. = ICMPv6 header (58 B) +
 * Byte 0: Cur Hop Limit (default TTL that should be used for outgoing)
 * Byte 1: 2 bit dhcpv6 flags, 6 reserved
 * Byte 2+3: Router lifetime (seconds, max. 9000).
 * Byte 4-7: Reachable time (ms)
 * Byte 8-11: Retrans timer (ms)
 * Byte 12 and rest: Options (there should be at least a prefix-info option) */
static void net_handlev6routeadv(uint16_t v6pllen) {
  uint8_t option; uint16_t optlen; uint16_t optpos;
  if (hdrbuf[21] != 255) { /* These requests MUST have a TTL of 255 */
    return;
  }
  if ((hdrbuf[60] == 0) && (hdrbuf[61] == 0)) {
    /* Router lifetime of 0 means: Do not use this as a default router */
    return;
  }
  /* Parse the options */
  optpos = 70;
  while ((optpos+3) < plen) { /* Room for another option */
    option = enc28j60_read_rx_memory(thispacketstart + optpos + 0);
    optlen = enc28j60_read_rx_memory(thispacketstart + optpos + 1);
    if ((option == 3) && (optlen == 4)) { /* Prefix information */
      uint8_t prefixlen; uint8_t rflags; uint8_t i;
      prefixlen = enc28j60_read_rx_memory(thispacketstart + optpos + 2);
      rflags = enc28j60_read_rx_memory(thispacketstart + optpos + 3);
      if ((prefixlen == 64) /* We cannot handle anything else */
       && ((rflags & 0x40) > 0)) { /* Valid for autoconfig */
          for (i = 0; i < 8; i++) {
            net_ip6[i] = enc28j60_read_rx_memory(thispacketstart + optpos + 16 + i);
          }
          net_filleui64(&net_ip6[8]);
      } else {
          console_printpgm_P(PSTR(" -EPREFIX "));
      }
    }
    optpos += optlen * 8;
  }
}

/* For reference: ICMPv6 header = IPv6 header (54 B) +
 * Byte 0: type; 135 = neighbour solicitation, XX = router advertisement
 * Byte 1: code
 * Byte 2+3: checksum */
static void net_handleicmp6(uint16_t v6pllen) {
  /* console_printpgm_P(PSTR(" ICMP type "));
  console_printhex8(hdrbuf[54]); */
  if ((hdrbuf[54] == 128) && (hdrbuf[55] == 0)) {
    net_handlev6ping(v6pllen);
  }
  if ((hdrbuf[54] == 134) && (hdrbuf[55] == 0)) {
    if (net_ipv6addrmode == 2) {
      net_handlev6routeadv(v6pllen);
    }
  }
  if ((hdrbuf[54] == 135) && (hdrbuf[55] == 0)) {
    net_handlev6neighboursol(v6pllen);
  }
}

/* for reference: IPv6 header = Ethernet header (14 B) +
 * Byte 0-3: 4 bit version (6), 8 bit traffic class, 20 bit flow label
 * Byte 4+5: payload length
 * Byte 6: next-header; values: 0x3a icmp6, 0x06 tcp, 0x11 UDP, 0x3b NONE
 * Byte 7: hop limit (=TTL)
 * Byte 8-23: Source Address
 * Byte 24-39: Destination Address
 */
static void net_handleip6(void) {
  uint16_t v6pllen;
  
  if ((hdrbuf[14] & 0xf0) != 0x60) {
    /* Not IPv6 */
    return;
  }
#if 0
  if (1) {
    uint16_t i; 
    console_printpgm_P(PSTR(" V6 Packet SRC "));
    for (i = 0; i < 16; i++) {
      console_printhex8(hdrbuf[14+8+i]);
    }
    console_printpgm_P(PSTR(" DST "));
    for (i = 0; i < 16; i++) {
      console_printhex8(hdrbuf[14+24+i]);
    }
    console_printpgm_P(PSTR(" NH "));
    console_printhex8(hdrbuf[14+6]);
    console_printpgm_P(PSTR(" l "));
    console_printhex8(hdrbuf[14+4]);
    console_printhex8(hdrbuf[14+5]);
    console_printpgm_P(PSTR("\r\n"));
  }
#endif /* lots of debug output */
  v6pllen = ((uint16_t)hdrbuf[14+4] << 8) | hdrbuf[14+5];
  if ((v6pllen + 54) > plen) { /* Payload length plus header length exceeds
                                * the length of the packet. Invalid packet. */
    console_printpgm_P(PSTR(" V6LenExceedsPacketLen "));
    return;
  }
  if ((memcmp(&hdrbuf[14+24], &net_llip6[0], 16) != 0)
   && (memcmp(&hdrbuf[14+24], &net_ip6[0], 16) != 0)) {
    /* OK, so this is not individually addressed to us, but maybe
     * to a valid multicast address? */
    if ((memcmp_P(&hdrbuf[14+24], ipv6_solicitednode, 13) !=0)
     && (memcmp_P(&hdrbuf[14+24], ipv6_allnodes, 16) != 0)) {
      /* Not addressed to one of our multicast addresses either. Not for us. */
      return;
    }
  }
  if (hdrbuf[14+6] == 0x3a) { /* nextheader = Protocol ICMPv6 */
    net_handleicmp6(v6pllen);
  } else if (hdrbuf[14+6] == 17) { /* nextheader = Protocol UDP */
    net_handleudp6();
  }
}

static void net_timeoutoldtcp(void) {
  uint8_t i;
  for (i = 0; i < NUMTCPCONNS; i++) {
    if (tcpconns[i].myport != 0) { /* this state entry is in use */
      if ((getticks() - tcpconns[i].lastactive) > (TCP_TIMEOUT * TICKSPERSECOND)) {
        /* Timed out */
        /* FIXME? We should actually send a TCP RST and stuff, but on the
         * other hand, lets not exaggerate */
        /* Mark as no longer used */
        ERRORconsole_pgm_P(PSTR(" TCP timeout "));
        tcpconns[i].myport = 0;
      }
    }
  }
}

void net_donetconsole(void) {
#ifdef NETCONSOLE
  uint8_t i;
  if ((console_opbnetpos > 0) && (txsending == 0)) {
    /* Hey, there is something to send and we're idle. */
    cli(); /* Unfortunately, we need to make sure the buffer doesn't get
            * modified in the mean time :-/ */
    /* Build packets */
    for (i = 0; i < NUMTCPCONNS; i++) {
      if ((tcpconns[i].myport == TELNETPORT)
       && (tcpconns[i].state == STATE_ESTABLISHED)) {
        /* This connection needs to get the data. */
        uint16_t txad;
        uint8_t txbnum;
        uint16_t j;
        txbnum = net_preparetxbuffer();
        txad = txbufstart[txbnum] + 1;
        /* Start of Ethernet header */
        for (j = 0; j < 6; j++) {
          enc28j60_write_memory(txad + 0 + j, tcpconns[i].partnermac[j]);
        }
        if (tcpconns[i].iptype == 0) { /* IPv4 */
          uint16_t iplen = 40 + console_opbnetpos;
          uint16_t cs;
          enc28j60_write_memory(txad + 12, 0x08); /* IP */
          enc28j60_write_memory(txad + 13, 0x00); /* IP */
          /* Start of IP header */
          enc28j60_write_memory(txad + 14, 0x45); /* IPv4 and 20 byte header length */
          enc28j60_write_memory(txad + 15, 0x10); /* diff serv. / TOS */
          enc28j60_write_memory(txad + 16, iplen >> 8); /* length */
          enc28j60_write_memory(txad + 17, iplen & 0xff); /* length */
          enc28j60_write_memory(txad + 18,   23); /* identification (nonsense) */
          enc28j60_write_memory(txad + 19, 0x23); /* identification */
          enc28j60_write_memory(txad + 20, 0x40); /* don't fragment, frag offset 0 */
          enc28j60_write_memory(txad + 21, 0x00); /* frag offset 0 */
          enc28j60_write_memory(txad + 22, NET_TTL); /* TTL */
          enc28j60_write_memory(txad + 23, 0x06); /* Protocol TCP */
          enc28j60_write_memory(txad + 24, 0x00); /* Checksum */
          enc28j60_write_memory(txad + 25, 0x00); /* Checksum */
          for (j = 0; j < 4; j++) {
            enc28j60_write_memory(txad + 26 + j, net_ip4[j]);
            enc28j60_write_memory(txad + 30 + j, tcpconns[i].partnerip4[j]);
          }
          cs = net_calcipchecksum(txad + 14, 20);
          enc28j60_write_memory(txad + 24, cs >> 8); /* Checksum */
          enc28j60_write_memory(txad + 25, cs & 0xff); /* Checksum */
          /* Start of TCP header */
          enc28j60_write_memory(txad + 34, tcpconns[i].myport >> 8);
          enc28j60_write_memory(txad + 35, tcpconns[i].myport & 0xff);
          enc28j60_write_memory(txad + 36, tcpconns[i].partnerport >> 8);
          enc28j60_write_memory(txad + 37, tcpconns[i].partnerport & 0xff);
          enc28j60_write_memory(txad + 38, tcpconns[i].mysnr >> 24);
          enc28j60_write_memory(txad + 39, tcpconns[i].mysnr >> 16);
          enc28j60_write_memory(txad + 40, tcpconns[i].mysnr >>  8);
          enc28j60_write_memory(txad + 41, tcpconns[i].mysnr >>  0);
          tcpconns[i].mysnr += console_opbnetpos;
          enc28j60_write_memory(txad + 42, tcpconns[i].partnersnr >> 24);
          enc28j60_write_memory(txad + 43, tcpconns[i].partnersnr >> 16);
          enc28j60_write_memory(txad + 44, tcpconns[i].partnersnr >>  8);
          enc28j60_write_memory(txad + 45, tcpconns[i].partnersnr >>  0);
          enc28j60_write_memory(txad + 46, 5 << 4); /* data offset = length of tcp header */
          enc28j60_write_memory(txad + 47, TCP_ACK | TCP_PSH);
          enc28j60_write_memory(txad + 48, 1500 >> 8);
          enc28j60_write_memory(txad + 49, 1500 & 0xff);
          enc28j60_write_memory(txad + 50, 0x00); /* Checksum */
          enc28j60_write_memory(txad + 51, 0x00); /* Checksum */
          enc28j60_write_memory(txad + 52, 0x00); /* Urgent */
          enc28j60_write_memory(txad + 53, 0x00); /* Urgent */
          /* Start of Data */
          for (j = 0; j < console_opbnetpos; j++) {
            enc28j60_write_memory(txad + 54 + j, console_opbufnet[j]);
          }
          /* Start of TCP pseudo header. We just put it behind the real packet. */
          j = txad + 54 + console_opbnetpos;
          if (console_opbnetpos & 1) { /* Alignment for the checksum */
            enc28j60_write_memory(j, 0);
            j++;
          }
          enc28j60_write_memory(j + 0, net_ip4[0]);
          enc28j60_write_memory(j + 1, net_ip4[1]);
          enc28j60_write_memory(j + 2, net_ip4[2]);
          enc28j60_write_memory(j + 3, net_ip4[3]);
          enc28j60_write_memory(j + 4, tcpconns[i].partnerip4[0]);
          enc28j60_write_memory(j + 5, tcpconns[i].partnerip4[1]);
          enc28j60_write_memory(j + 6, tcpconns[i].partnerip4[2]);
          enc28j60_write_memory(j + 7, tcpconns[i].partnerip4[3]);
          enc28j60_write_memory(j + 8, 0); /* just 0. */
          enc28j60_write_memory(j + 9, 0x06); /* Prot. number TCP */
          enc28j60_write_memory(j + 10, (20 + console_opbnetpos) >> 8); /* length */
          enc28j60_write_memory(j + 11, (20 + console_opbnetpos) & 0xff); /* length */
          /* Now fix the checksum */
          if (console_opbnetpos & 1) {
            cs = net_calcipchecksum(txad + 34, 20 + console_opbnetpos + 13);
          } else {
            cs = net_calcipchecksum(txad + 34, 20 + console_opbnetpos + 12);
          }
          enc28j60_write_memory(txad + 50, cs >> 8); /* Checksum */
          enc28j60_write_memory(txad + 51, cs & 0xff); /* Checksum */
          txbufpsize[txbnum] = 54 + console_opbnetpos;
          /* ready to be sent */
        }
      }
    }
    console_opbnetpos = 0; /* Free buffer */
    sei();
  }
#endif /* NETCONSOLE */
}

/* shows active connections to the console. called from there
 * WITH INTERRUPTS DISABLED! */
void net_showconnections(void) {
#if ((defined(SERIALCONSOLE)) || (defined(NETCONSOLE)))
  uint8_t i;
  uint8_t a = 0;
  console_printpgm_noirq_P(PSTR("RemoteIP                         RPrt LPrt Last\r\n"));
  for (i = 0; i < NUMTCPCONNS; i++) {
    if (tcpconns[i].myport != 0) { /* this state entry is in use */
      uint8_t j; uint16_t lastact;
      a++;
      if (tcpconns[i].iptype == 0) { /* V4 */
        for (j = 0; j < 4; j++) {
          console_printdec_noirq(tcpconns[i].partnerip4[j]);
          if (j != 3) {
            console_printpgm_noirq_P(PSTR("."));
          }
        }
        console_printpgm_noirq_P(PSTR("                 "));
      } else { /* V6 */
        for (j = 0; j < 16; j++) {
          console_printhex8_noirq(tcpconns[i].partnerip6[j]);
        }
      }
      console_printpgm_P(PSTR(" "));
      console_printhex8_noirq(tcpconns[i].partnerport >> 8);
      console_printhex8_noirq(tcpconns[i].partnerport >> 0);
      console_printpgm_P(PSTR(" "));
      console_printhex8_noirq(tcpconns[i].myport >> 8);
      console_printhex8_noirq(tcpconns[i].myport >> 0);
      lastact = getticks() - tcpconns[i].lastactive;
      console_printpgm_P(PSTR(" "));
      console_printhex8_noirq(lastact >> 8);
      console_printhex8_noirq(lastact >> 0);
    }
  }
  console_printpgm_noirq_P(PSTR("\r\n"));
  console_printdec_noirq(a);
  console_printpgm_noirq_P(PSTR(" active connections."));
#endif  /*  ((defined(SERIALCONSOLE)) || (defined(NETCONSOLE))) */
}

/* Initialization function - inits the network chip and internal variables */
void enc28j60_init(void) {
  uint8_t i;
  /* FIXME SPI powersave aus? */
  /* sollte default 0 sein. */
  
  /* Set MOSI, SCK and SS output, MISO input */
  DDRB |= _BV(PB3) | _BV(PB5) | _BV(PB2);
  DDRB &= (uint8_t)~_BV(PB4);
  spi_ss_hi();
  /* Enable SPI, Master, set clock rate fck/4 */
  SPCR |= _BV(SPE) | _BV(MSTR);
  /* Now double clockrate */
  SPSR |= _BV(SPI2X);
  
  /* Send reset command */
  spi_ss_lo();
  spi_sendbyte(0xff);
  spi_ss_hi();

  /* WAIT */
  _delay_ms(500.0);

  /* Initialize */
  /* Set receive buffer area */
  enc28j60_write_cont_reg(E28J60_ERXSTL, 0);
  enc28j60_write_cont_reg(E28J60_ERXSTH, 0);
  enc28j60_write_cont_reg(E28J60_ERXNDL, (RXBUFEND & 0xff));
  enc28j60_write_cont_reg(E28J60_ERXNDH, (RXBUFEND >> 8));
  /* Also set receive buffer pointer. */
  enc28j60_write_cont_reg(E28J60_ERXRDPTL, 0);
  enc28j60_write_cont_reg(E28J60_ERXRDPTH, 0);
  /* TX Buffer gets set with every packet sent, so there is nothing to
   * program on the NIC. */
  for (i = 0; i < NUMTXBUFS; i++) {
    txbufused[i] = 0;
    txbufstart[i] = RXBUFEND + 1 + i * 1526;
  }
  txsending = 0;
  /* Set write and read pointers so they match our internal variables */
  currxad = curtxad = 0;
  enc28j60_write_cont_reg(E28J60_ERDPTL, 0);
  enc28j60_write_cont_reg(E28J60_ERDPTH, 0);
  enc28j60_write_cont_reg(E28J60_EWRPTL, 0);
  enc28j60_write_cont_reg(E28J60_EWRPTH, 0);
  
  /* Filter configuration: accept unicast, multicast, broadcast, and check CRCs. */
  enc28j60_write_cont_reg(E28J60_ERXFCON, E28J60_ERXFCON_UCEN
                                        | E28J60_ERXFCON_CRCEN
                                        | E28J60_ERXFCON_MCEN
                                        | E28J60_ERXFCON_BCEN);
  
  /* prepare MAC */
  /* enable MAC receive and flow control. */
  enc28j60_write_cont_reg(E28J60_MACON1, E28J60_MACON1_MARXEN
                                       | E28J60_MACON1_RXPAUS
                                       | E28J60_MACON1_TXPAUS);
  /* This whole register is UNDOCUMENTED in the current documentation, but
   * old code often references it. In old code comments usually claim that
   * it is necessary to set this to 0 in order to take the MAC out of the
   * reset. However, this really does not seem to be necessary - at least
   * with the chip revision I have. */
  /* enc28j60_write_cont_reg(E28J60_MACON2, 0); */
  /* Pad short frames to 60 bytes and automatically add CRC */
  enc28j60_write_cont_reg(E28J60_MACON3, E28J60_MACON3_TXCRCEN
                                       | E28J60_MACON3_PADCFG0
                                       | ((net_fdx) ? E28J60_MACON3_FULDPX : 0));
  /* enable transmit defers (standard requires it?) */
  enc28j60_write_cont_reg(E28J60_MACON4, E28J60_MACON4_DEFER);
  /* set maximum frame length */
  enc28j60_write_cont_reg(E28J60_MAMXFLL, 1518 & 0xff);
  enc28j60_write_cont_reg(E28J60_MAMXFLL, 1518 >> 8);
  /* Back-to-back inter-packet gap. According to manual: 0x12 for half, 0x15 for full duplex */
  enc28j60_write_cont_reg(E28J60_MABBIPG, (net_fdx) ? 0x15 : 0x12);
  /* non-back-to-back inter-packet gap low. value from manual. */
  enc28j60_write_cont_reg(E28J60_MAIPGL, 0x12);
  /* non-back-to-back inter-packet gap high. manual: 0x0c for half duplex */
  enc28j60_write_cont_reg(E28J60_MAIPGH, 0x0c);
  /* configure MAC address. */
  enc28j60_write_cont_reg(E28J60_MAADR1, net_mac[0]);
  enc28j60_write_cont_reg(E28J60_MAADR2, net_mac[1]);
  enc28j60_write_cont_reg(E28J60_MAADR3, net_mac[2]);
  enc28j60_write_cont_reg(E28J60_MAADR4, net_mac[3]);
  enc28j60_write_cont_reg(E28J60_MAADR5, net_mac[4]);
  enc28j60_write_cont_reg(E28J60_MAADR6, net_mac[5]);
  /* Disable clock output to clkout pin. */
  enc28j60_write_cont_reg(E28J60_ECOCON, 0);
  
  /* PHY config */
  /* set half/full duplex */
  enc28j60_write_phy(E28J60_PHCON1, (net_fdx) ? E28J60_PHCON1_PDPXMD : 0);
  /* disable loopback */
  enc28j60_write_phy(E28J60_PHCON2, E28J60_PHCON2_HDLDIS);
  
  /* Enable automatic pointer increment */
  enc28j60_bitfield_set(E28J60_ECON2, E28J60_ECON2_AUTOINC);
  /* Enable reception in ECON1 */
  enc28j60_bitfield_set(E28J60_ECON1, E28J60_ECON1_RXEN);
  
  if (net_ipv6addrmode > 0) { /* IPv6 not disabled */
    /* Initialize our link local IPv6 address */
    net_llip6[0] = 0xfe; net_llip6[1] = 0x80;
    for (i = 2; i < 8; i++) {
      net_llip6[i] = 0;
    }
    net_filleui64(&net_llip6[8]);
  }
  
  /* Enable the network IRQ (IRQ1) and let it trigger on the falling edge. */
  EICRA |= _BV(ISC11);
  EIMSK |= _BV(INT1);
  /* Tell the NIC to generate IRQs */
  enc28j60_write_cont_reg(E28J60_EIE,
                          (E28J60_EIE_INTIE | E28J60_EIE_PKTIE | E28J60_EIE_TXIE));
}

void net_worker(void) {
  uint8_t i;
  /* check if a packet is ready to be received */
  if (enc28j60_read_cont_reg(E28J60_EPKTCNT)) {
    uint16_t nextpacketstart;
    thispacketstart = (enc28j60_read_cont_reg(E28J60_ERXRDPTH) << 8)
                    | enc28j60_read_cont_reg(E28J60_ERXRDPTL);
    nextpacketstart = enc28j60_read_rx_memory(thispacketstart)
                    | (enc28j60_read_rx_memory(thispacketstart+1) << 8);
    plen = enc28j60_read_rx_memory(thispacketstart+2)
         | (enc28j60_read_rx_memory(thispacketstart+3) << 8);
#if 0
    console_printpgm_P(PSTR("\r\nPR: st "));
    console_printhex8(thispacketstart >> 8);
    console_printhex8(thispacketstart & 0xff);
    console_printpgm_P(PSTR(" ne "));
    console_printhex8(nextpacketstart >> 8);
    console_printhex8(nextpacketstart & 0xff);
#endif /* 0 */
    /* possibly check some of the receive status bits here? */
    if ((plen < 64) || (plen > 1518)) { /* too short / long. invalid packet. */
      console_printpgm_P(PSTR(" inv plen "));
      goto getout;
    }
    plen -= 4; /* Discard the CRC at the end */
    thispacketstart += 6; /* Advance to start of ethernet frame */
    /* Copy the whole header */
    for (i = 0; i < HEADERBUFSIZE; i++) {
      if (i < plen) {
        hdrbuf[i] = enc28j60_read_rx_memory(thispacketstart + i);
      } else {
        hdrbuf[i] = 0; /* Make sure we cannot use contents from the previous
                        * packet accidently */
      }
    }
    /* For reference: Structure of an Ethernet Packet:
     * Destination          Byte  0 -  5
     * Source               Byte  6 - 11
     * Length/Type          Byte 12 - 13
     * Data                 at least 46 Bytes */
    /* Sanity check, so that we don't reply to
     * multi-/broadcast-addresses and create a flood */
    if ((hdrbuf[6] & NET_MCASTBIT) != 0) {
      console_printpgm_P(PSTR(" brc src "));
      goto getout;
    }
    /* OK, so is it addressed to us? Actually the NIC should already
     * check that, but better safe than sorry. */
    if (memcmp(&hdrbuf[0], &net_mac[0], 6) != 0) { /* not adressed to us? */
      if ((hdrbuf[0] & NET_MCASTBIT) == 0x00) { /* not broad- or multi-cast? */
        console_printpgm_P(PSTR(" pnfu "));
        goto getout;
      }
    }
    /* handle packet */
    if ((hdrbuf[12] == 0x08) /* Protocol ARP */
     && (hdrbuf[13] == 0x06)) {
      net_handlearp();
    } else if ((hdrbuf[12] == 0x08) /* Protocol IP */
            && (hdrbuf[13] == 0x00)) {
      net_handleip4();
    } else if ((hdrbuf[12] == 0x86) /* Protocol IPv6 */
            && (hdrbuf[13] == 0xdd)) {
      if (net_ipv6addrmode > 0) {
        net_handleip6();
      }
    }
getout:
    /* now advance the packet pointer to free ring buffer memory */
    enc28j60_write_cont_reg(E28J60_ERXRDPTL, nextpacketstart & 0xff);
    enc28j60_write_cont_reg(E28J60_ERXRDPTH, nextpacketstart >> 8);
    /* And tell the NIC to decrese the packet count */
    enc28j60_bitfield_set(E28J60_ECON2, E28J60_ECON2_PKTDEC);
  }
  /* Check if a transmit completed */
  if (txsending > 0) {
    if ((enc28j60_read_cont_reg(E28J60_ECON1) & E28J60_ECON1_TXRTS) == 0) {
//       console_printpgm_P(PSTR(" txc "));
      txbufused[txsending - 1] = 0;
      txsending = 0;
    }
  }
  net_timeoutoldtcp();
  net_trysending();
  /* check netconsole and send */
  net_donetconsole();
  /* check if there are things to send */
  net_trysending();
}
