/* $Id: timers.c,v 1.2 2010-08-22 07:23:57 simimeie Exp $
 * Functions for timing.
 * This mainly allows you to get a timestamp value - a very exact one
 * (resolution 1 clock cycle) or a less exact one.
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include "timers.h"

/* Internal ticks */
static volatile uint16_t ticks = 0;
static volatile uint16_t tovfls = 0;

/* Executes every 65536 cpu cycles when the timer overflows */
ISR(TIMER1_OVF_vect)
{
  if (ticks == 0xffff) {
    tovfls++;
    ticks = 0;
  } else {
    ticks++;
  }
}

uint16_t getticks(void)
{
  uint16_t res;
  cli();
  res = ticks;
  sei();
  return res;
}

uint16_t getticks_noirq(void)
{
  return ticks;
}

void gettimestamp(struct timestamp * ts)
{
  uint16_t cou1, cou2, tick1;
  ts->tovfls = tovfls;
  /* Low byte must be read first for 16bit registers */
  cou1 = TCNT1L;
  cou1 |= (uint16_t)(TCNT1H) << 8;
  tick1 = getticks(); /* enables interrupts! */
  cou2 = TCNT1L;
  cou2 |= (uint16_t)(TCNT1H) << 8;
  if (cou1 > cou2) { /* An overflow occured in between */
    /* get new ticks value - we could already have it, but we do not know
     * for sure. If we get another one now, we do - because there is no
     * way we're going to need another 65k cpu cycles. */
    ts->tovfls = tovfls;
    ts->ticks = getticks();
  } else { /* no overflow. */
    ts->ticks = tick1;
  }
  ts->partticks = cou2;
}

void gettimestamp_noirq(struct timestamp * ts)
{
  uint16_t cou1, cou2; uint8_t ovf;
  /* Low byte must be read first for 16bit registers */
  cou1 = TCNT1L;
  cou1 |= (uint16_t)(TCNT1H) << 8;
  /* Check if there is an overflow pending */
  ovf = TIFR1 & _BV(TOV1);
  if (ovf) {
    /* Overflow pending, reread the counter to be sure */
    cou2 = TCNT1L;
    cou2 |= (uint16_t)(TCNT1H) << 8;
    if (ticks == 0xffff) {
      ts->tovfls = tovfls + 1;
    } else {
      ts->tovfls = tovfls;
    }
    ts->ticks = ticks + 1;
    ts->partticks = cou2;
  } else { /* No overflow is pending */
    ts->tovfls = tovfls;
    ts->ticks = ticks;
    ts->partticks = cou1;
  }
}

void timers_init(void)
{
  /* Enable counter1, no prescaling (1 increment per clock cycle!) */
  TCCR1B |= _BV(CS10);
  /* Enable interrupt when timer overflows (at 65536) */
  TIMSK1 |= _BV(TOIE1);
}
